import asyncio
import logging
import mimetypes
import tarfile
import tempfile
import uuid
from collections.abc import AsyncIterator, Callable, MutableMapping
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Final, Optional, cast, override

import aiofiles
import aiohttp

from ai.backend.common.artifact_storage import AbstractStorage, AbstractStoragePool
from ai.backend.common.bgtask.bgtask import BackgroundTaskManager, ProgressReporter
from ai.backend.common.clients.valkey_client.valkey_artifact.client import (
    ValkeyArtifactDownloadTrackingClient,
)
from ai.backend.common.data.artifact.types import ArtifactRegistryType
from ai.backend.common.data.storage.registries.types import FileObjectData, ModelTarget
from ai.backend.common.data.storage.types import (
    ArtifactStorageImportStep,
)
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.events.event_types.artifact.anycast import ModelImportDoneEvent
from ai.backend.common.types import DispatchResult, StreamReader
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.storage.client.manager import ManagerHTTPClient, ManagerHTTPClientArgs
from ai.backend.storage.client.s3 import S3Client
from ai.backend.storage.config.unified import (
    ReservoirClientConfig,
    ReservoirConfig,
)
from ai.backend.storage.context_types import ArtifactVerifierContext
from ai.backend.storage.exception import (
    ArtifactRevisionEmptyError,
    ArtifactStorageEmptyError,
    ObjectStorageBucketNotFoundError,
    ReservoirStorageConfigInvalidError,
    StorageNotFoundError,
    StorageStepRequiredStepNotProvided,
)
from ai.backend.storage.services.artifacts.common import ModelArchiveStep, ModelVerifyStep
from ai.backend.storage.services.artifacts.storage_transfer import StorageTransferManager
from ai.backend.storage.services.artifacts.types import (
    DownloadStepResult,
    ImportPipeline,
    ImportStep,
    ImportStepContext,
)
from ai.backend.storage.storages.object_storage import ObjectStorage
from ai.backend.storage.storages.storage_pool import StoragePool
from ai.backend.storage.storages.vfs_storage import VFSStorage
from ai.backend.storage.types import BucketCopyOptions

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

_DOWNLOAD_PROGRESS_UPDATE_INTERVAL: Final[int] = 30


class ReservoirVFSDownloadStreamReader(StreamReader):
    """StreamReader that wraps ManagerHTTPClient VFS download stream for individual files."""

    _client: ManagerHTTPClient
    _storage_name: str
    _filepath: str

    def __init__(
        self,
        client: ManagerHTTPClient,
        storage_name: str,
        filepath: str,
    ):
        self._client = client
        self._storage_name = storage_name
        self._filepath = filepath

    @override
    async def read(self) -> AsyncIterator[bytes]:
        async for chunk in self._client.download_vfs_file_streaming(
            self._storage_name, self._filepath
        ):
            yield chunk

    @override
    def content_type(self) -> Optional[str]:
        # Guess content type from file extension
        content_type, _ = mimetypes.guess_type(self._filepath)
        return content_type or "application/octet-stream"


class ReservoirVFSFileDownloader:
    """Helper class to download individual files from VFS storage to local filesystem."""

    _client: ManagerHTTPClient
    _storage_name: str
    _redis_client: ValkeyArtifactDownloadTrackingClient
    _model_id: str
    _revision: str
    _download_complete: bool
    _progress_task: Optional[asyncio.Task[None]]
    _bytes_downloaded: int

    def __init__(
        self,
        client: ManagerHTTPClient,
        storage_name: str,
        redis_client: ValkeyArtifactDownloadTrackingClient,
        model_id: str,
        revision: str,
    ):
        self._client = client
        self._storage_name = storage_name
        self._redis_client = redis_client
        self._model_id = model_id
        self._revision = revision
        self._download_complete = False
        self._progress_task = None
        self._bytes_downloaded = 0

    async def _periodic_progress_update(
        self,
        total_bytes: int,
        file_path: str,
    ) -> None:
        """
        Background task that periodically updates Redis with download progress.

        :param total_bytes: Total bytes for this file
        :param file_path: Path of the file being downloaded
        """
        while not self._download_complete:
            await asyncio.sleep(_DOWNLOAD_PROGRESS_UPDATE_INTERVAL)

            try:
                await self._redis_client.update_file_progress(
                    model_id=self._model_id,
                    revision=self._revision,
                    file_path=file_path,
                    current_bytes=self._bytes_downloaded,
                    total_bytes=total_bytes,
                    success=False,
                )
            except asyncio.CancelledError:
                # Task is being cancelled, exit cleanly
                break
            except Exception as e:
                # Log error but don't fail the download
                log.warning(
                    "Failed to update download progress in Redis: {}",
                    str(e),
                )

    async def download_file(self, remote_path: str, local_path: Path, total_bytes: int) -> int:
        """
        Download a single file from VFS storage to local filesystem.

        Args:
            remote_path: Path of the file in VFS storage
            local_path: Local filesystem path to save the file
            total_bytes: Total size of the file in bytes

        Returns:
            Number of bytes downloaded
        """
        # Ensure parent directory exists
        local_path.parent.mkdir(parents=True, exist_ok=True)

        # Initialize download state
        self._bytes_downloaded = 0
        self._download_complete = False
        self._progress_task = None

        stream_reader = ReservoirVFSDownloadStreamReader(
            self._client, self._storage_name, remote_path
        )

        try:
            # Start background progress task
            self._progress_task = asyncio.create_task(
                self._periodic_progress_update(
                    total_bytes=total_bytes,
                    file_path=remote_path,
                )
            )

            async with aiofiles.open(local_path, "wb") as f:
                try:
                    async for chunk in stream_reader.read():
                        await f.write(chunk)
                        self._bytes_downloaded += len(chunk)
                except aiohttp.ClientError as e:
                    log.error(
                        f"Network error during download: {e}, Downloaded {self._bytes_downloaded} bytes before failure"
                    )
                    raise
                except asyncio.TimeoutError:
                    log.error(f"Timeout after downloading {self._bytes_downloaded} bytes")
                    raise
        except Exception as e:
            # Update Redis with error status for any unexpected errors
            try:
                await self._redis_client.update_file_progress(
                    model_id=self._model_id,
                    revision=self._revision,
                    file_path=remote_path,
                    current_bytes=self._bytes_downloaded,
                    total_bytes=total_bytes,
                    success=False,
                    error_message=str(e),
                )
            except Exception as redis_err:
                log.warning(f"Failed to update error status in Redis: {redis_err}")
            raise
        finally:
            self._download_complete = True

            # Cancel and wait for progress task
            if self._progress_task:
                self._progress_task.cancel()
                try:
                    await self._progress_task
                except asyncio.CancelledError:
                    pass

            # Final update to Redis with success status
            try:
                await self._redis_client.update_file_progress(
                    model_id=self._model_id,
                    revision=self._revision,
                    file_path=remote_path,
                    current_bytes=self._bytes_downloaded,
                    total_bytes=total_bytes,
                    success=(self._bytes_downloaded >= total_bytes),
                )
            except Exception as redis_err:
                log.warning(f"Failed to update final status in Redis: {redis_err}")

        log.debug(
            f"Downloaded file: {remote_path} -> {local_path} ({self._bytes_downloaded} bytes)"
        )
        return self._bytes_downloaded


class ReservoirS3FileDownloadStreamReader(StreamReader):
    _src_s3_client: S3Client
    _key: str
    _size: int
    _options: BucketCopyOptions
    _download_chunk_size: int
    _content_type: Optional[str]
    _redis_client: ValkeyArtifactDownloadTrackingClient
    _model_id: str
    _revision: str
    _download_complete: bool
    _progress_task: Optional[asyncio.Task[None]]

    def __init__(
        self,
        src_s3_client: S3Client,
        key: str,
        size: int,
        options: BucketCopyOptions,
        download_chunk_size: int,
        content_type: Optional[str],
        redis_client: ValkeyArtifactDownloadTrackingClient,
        model_id: str,
        revision: str,
    ):
        self._src_s3_client = src_s3_client
        self._key = key
        self._size = size
        self._options = options
        self._download_chunk_size = download_chunk_size
        self._content_type = content_type
        self._redis_client = redis_client
        self._model_id = model_id
        self._revision = revision
        self._download_complete = False
        self._progress_task = None

    async def _periodic_progress_update(
        self,
        offset_getter: Callable[[], int],
        total_bytes: int,
    ) -> None:
        """
        Background task that periodically updates Redis with download progress.

        :param offset_getter: Callable that returns current download offset
        :param total_bytes: Total bytes for this file
        """
        while not self._download_complete:
            await asyncio.sleep(_DOWNLOAD_PROGRESS_UPDATE_INTERVAL)

            try:
                current = offset_getter()
                await self._redis_client.update_file_progress(
                    model_id=self._model_id,
                    revision=self._revision,
                    file_path=self._key,
                    current_bytes=current,
                    total_bytes=total_bytes,
                    success=False,
                )
            except asyncio.CancelledError:
                # Task is being cancelled, exit cleanly
                break
            except Exception as e:
                # Log error but don't fail the download
                log.warning(
                    "Failed to update download progress in Redis: {}",
                    str(e),
                )

    @override
    async def read(self) -> AsyncIterator[bytes]:
        sent = 0
        next_mark = self._options.progress_log_interval_bytes
        download_stream = self._src_s3_client.download_stream(
            self._key, chunk_size=self._download_chunk_size
        )

        self._download_complete = False
        self._progress_task = None

        try:
            # Start background progress task
            self._progress_task = asyncio.create_task(
                self._periodic_progress_update(
                    offset_getter=lambda: sent,
                    total_bytes=self._size,
                )
            )

            async for chunk in download_stream.read():
                sent += len(chunk)
                if next_mark and sent >= next_mark:
                    log.trace(
                        "[stream_bucket_to_bucket] progress key={} sent={}/{}",
                        self._key,
                        sent,
                        self._size,
                    )
                    next_mark += self._options.progress_log_interval_bytes

                yield chunk
        except Exception as e:
            # Update Redis with error status
            try:
                await self._redis_client.update_file_progress(
                    model_id=self._model_id,
                    revision=self._revision,
                    file_path=self._key,
                    current_bytes=sent,
                    total_bytes=self._size,
                    success=False,
                    error_message=str(e),
                )
            except Exception as redis_err:
                log.warning(f"Failed to update error status in Redis: {redis_err}")
            raise
        finally:
            self._download_complete = True

            # Cancel and wait for progress task
            if self._progress_task:
                self._progress_task.cancel()
                try:
                    await self._progress_task
                except asyncio.CancelledError:
                    pass

            # Final update to Redis with success status
            try:
                await self._redis_client.update_file_progress(
                    model_id=self._model_id,
                    revision=self._revision,
                    file_path=self._key,
                    current_bytes=sent,
                    total_bytes=self._size,
                    success=(sent >= self._size),
                )
            except Exception as redis_err:
                log.warning(f"Failed to update final status in Redis: {redis_err}")

    @override
    def content_type(self) -> Optional[str]:
        return self._content_type


@dataclass
class ReservoirServiceArgs:
    background_task_manager: BackgroundTaskManager
    event_producer: EventProducer
    storage_pool: StoragePool
    reservoir_registry_configs: dict[str, ReservoirConfig]
    artifact_verifier_ctx: ArtifactVerifierContext
    manager_http_clients: MutableMapping[str, ManagerHTTPClient]
    reservoir_client_config: ReservoirClientConfig  # Passed to ImportStepContext
    redis_client: ValkeyArtifactDownloadTrackingClient


class ReservoirService:
    """Service for Reservoir model operations"""

    _background_task_manager: BackgroundTaskManager
    _event_producer: EventProducer
    _reservoir_registry_configs: dict[str, ReservoirConfig]
    _storage_pool: StoragePool
    _transfer_manager: StorageTransferManager
    _artifact_verifier_ctx: ArtifactVerifierContext
    _manager_http_clients: MutableMapping[str, ManagerHTTPClient]
    _reservoir_client_config: ReservoirClientConfig
    _redis_client: ValkeyArtifactDownloadTrackingClient

    def __init__(self, args: ReservoirServiceArgs):
        self._background_task_manager = args.background_task_manager
        self._event_producer = args.event_producer
        self._reservoir_registry_configs = args.reservoir_registry_configs
        self._storage_pool = args.storage_pool
        self._transfer_manager = StorageTransferManager(args.storage_pool)
        self._artifact_verifier_ctx = args.artifact_verifier_ctx
        self._manager_http_clients = args.manager_http_clients
        self._reservoir_client_config = args.reservoir_client_config
        self._redis_client = args.redis_client

    async def import_model(
        self,
        registry_name: str,
        model: ModelTarget,
        reporter: ProgressReporter,
        storage_step_mappings: dict[ArtifactStorageImportStep, str],
        pipeline: ImportPipeline,
    ) -> None:
        """
        Import a single model from a reservoir registry to a reservoir storage.

        Args:
            registry_name: Name of the Reservoir registry
            model: Reservoir model to import
            reporter: ProgressReporter for tracking progress
            storage_step_mappings: Mapping of import steps to storage names
            pipeline: ImportPipeline to execute
        """
        success = False
        try:
            if model.revision is None:
                raise ArtifactRevisionEmptyError(f"Revision must be specified for model: {model}")

            # Create import context
            context = ImportStepContext(
                model=model,
                registry_name=registry_name,
                storage_pool=self._storage_pool,
                storage_step_mappings=storage_step_mappings,
                step_metadata={},
            )

            # Execute import pipeline
            await pipeline.execute(context)
            log.info(f"Model import completed: {model}")
            success = True
        finally:
            await self._event_producer.anycast_event(
                ModelImportDoneEvent(
                    success=success,
                    model_id=model.model_id,
                    revision=model.resolve_revision(ArtifactRegistryType.RESERVOIR),
                    registry_name=registry_name,
                    registry_type=ArtifactRegistryType.RESERVOIR,
                    # Reservoir registry's artifact's digest will be synced through scan API later
                    digest=None,
                )
            )

    async def import_models_batch(
        self,
        registry_name: str,
        models: list[ModelTarget],
        storage_step_mappings: dict[ArtifactStorageImportStep, str],
        pipeline: ImportPipeline,
    ) -> uuid.UUID:
        async def _import_models_batch(reporter: ProgressReporter) -> DispatchResult:
            model_count = len(models)
            if not model_count:
                log.warning("No models to import")
                return DispatchResult.error("No models provided for batch import")

            reporter.total_progress = model_count

            log.info(f"Starting batch model import: model_count={model_count}")

            try:
                successful_models = 0
                failed_models = 0
                errors = []

                # Process each model sequentially to avoid overwhelming the system
                # In a production system, this could be enhanced with parallel processing
                # and proper job queue management
                for idx, model in enumerate(models, 1):
                    model_id = model.model_id
                    try:
                        log.info(
                            f"Processing model in batch: model_id={model_id}, progress={idx}/{model_count}"
                        )

                        # TODO: Batch import logic can be optimized further
                        await self.import_model(
                            registry_name=registry_name,
                            model=model,
                            reporter=reporter,
                            storage_step_mappings=storage_step_mappings,
                            pipeline=pipeline,
                        )

                        successful_models += 1
                        log.info(
                            f"Successfully imported model in batch: model_id={model_id}, progress={idx}/{model_count}"
                        )
                    except Exception as e:
                        failed_models += 1
                        log.error(
                            f"Failed to import model in batch: {str(e)}, model_id={model_id}, progress={idx}/{model_count}"
                        )
                        errors.append(str(e))
                    finally:
                        await reporter.update(
                            1,
                            message=f"Processed model: {model_id} (progress: {idx}/{model_count})",
                        )

                log.info(
                    f"Batch model import completed: total_models={model_count}, "
                    f"successful_models={successful_models}, failed_models={failed_models}"
                )

                if failed_models > 0:
                    log.warning(
                        f"Some models failed to import in batch: failed_count={failed_models}"
                    )
                    return DispatchResult.partial_success(None, errors=errors)
            except Exception as e:
                log.error(f"Batch model import failed: {str(e)}")
                return DispatchResult.error(f"Batch import failed: {str(e)}")

            return DispatchResult.success(None)

        bgtask_id = await self._background_task_manager.start(_import_models_batch)
        return bgtask_id


# Import Pipeline Steps


class ReservoirDownloadStep(ImportStep[None]):
    """Step to copy files from Reservoir (effectively direct copy to download storage)"""

    def __init__(
        self,
        registry_configs: dict[str, ReservoirConfig],
        download_storage: AbstractStorage,
        manager_http_clients: MutableMapping[str, ManagerHTTPClient],
        reservoir_client_config: ReservoirClientConfig,
        redis_client: ValkeyArtifactDownloadTrackingClient,
    ) -> None:
        self._registry_configs = registry_configs
        self._download_storage = download_storage
        self._manager_http_clients = manager_http_clients
        self._reservoir_client_config = reservoir_client_config
        self._redis_client = redis_client

    @property
    def step_type(self) -> ArtifactStorageImportStep:
        return ArtifactStorageImportStep.DOWNLOAD

    @override
    async def execute(self, context: ImportStepContext, input_data: None) -> DownloadStepResult:
        # Get storage mapping for download step
        download_storage_name = context.storage_step_mappings.get(
            ArtifactStorageImportStep.DOWNLOAD
        )
        if not download_storage_name:
            raise StorageStepRequiredStepNotProvided("Download storage not specified in mappings")

        # Get registry configuration
        registry_config = self._registry_configs.get(context.registry_name)
        if not registry_config:
            raise ReservoirStorageConfigInvalidError(
                f"Registry configuration not found for: {context.registry_name}"
            )

        # Get the registry record to check storage_name
        # For now, we'll assume the storage_name from the registry configuration
        # In a real implementation, you'd query the database to get this information
        model = context.model
        revision = model.resolve_revision(ArtifactRegistryType.RESERVOIR)
        model_prefix = f"{model.model_id}/{revision}"

        # Determine storage type based on the actual download storage object type
        if isinstance(self._download_storage, VFSStorage):
            storage_type = "vfs"
        elif isinstance(self._download_storage, ObjectStorage):
            storage_type = "object_storage"
        else:
            storage_type = "unknown"

        bytes_copied = 0
        downloaded_files: list[tuple[FileObjectData, str]] = []

        if storage_type == "vfs":
            # Handle VFS storage type
            dest_path = (
                cast(VFSStorage, self._download_storage).base_path / model.model_id / revision
            )
            bytes_copied = await self._handle_vfs_download(
                registry_config, context, model_prefix, dest_path, self._reservoir_client_config
            )
            # For VFS downloads, create a single entry representing the extracted archive
            file_obj = FileObjectData(
                path=str(dest_path),
                size=bytes_copied,
                type="directory",
                download_url="",  # Not applicable for VFS
            )
            downloaded_files.append((file_obj, model_prefix))
        # TODO: This only make sense when both storage are ObjectStorage
        elif storage_type == "object_storage":
            # Handle object storage type
            downloaded_files, bytes_copied = await self._handle_object_storage_download(
                registry_config, download_storage_name, context, model_prefix
            )
        else:
            raise ReservoirStorageConfigInvalidError(
                f"Unsupported storage type: {storage_type} (storage class: {type(self._download_storage)})"
            )

        log.info(f"Reservoir copy completed: {context.model}, bytes_copied={bytes_copied}")

        return DownloadStepResult(
            downloaded_files=downloaded_files,
            storage_name=download_storage_name,  # Downloaded to download storage
            total_bytes=bytes_copied,
        )

    async def _handle_vfs_download(
        self,
        registry_config: ReservoirConfig,
        context: ImportStepContext,
        model_prefix: str,
        dest_path: Path,
        reservoir_client_config: ReservoirClientConfig,
    ) -> int:
        """Handle file downloads for VFS storage type using individual file downloads."""

        # Use the pre-resolved download storage object
        storage = self._download_storage
        if not isinstance(storage, VFSStorage):
            raise StorageNotFoundError(
                f"Download storage is not a VFS storage type: {type(storage)}"
            )

        try:
            if (
                not registry_config.manager_endpoint
                or not registry_config.manager_access_key
                or not registry_config.manager_secret_key
                or not registry_config.manager_api_version
            ):
                raise ReservoirStorageConfigInvalidError(
                    f"Manager access key not configured for reservoir registry: {context.registry_name}"
                )

            # Get or create ManagerHTTPClient from the pool
            if context.registry_name not in self._manager_http_clients:
                manager_client = ManagerHTTPClient(
                    ManagerHTTPClientArgs(
                        name=context.registry_name,
                        endpoint=registry_config.manager_endpoint,
                        access_key=registry_config.manager_access_key,
                        secret_key=registry_config.manager_secret_key,
                        api_version=registry_config.manager_api_version,
                        client_config=reservoir_client_config,
                    )
                )
                self._manager_http_clients[context.registry_name] = manager_client
            else:
                manager_client = self._manager_http_clients[context.registry_name]

            if not registry_config.storage_name:
                raise ReservoirStorageConfigInvalidError(
                    f"Reservoir registry storage name not configured: {context.registry_name}"
                )

            # Get list of all files in the model directory
            log.debug(f"Listing files for model: {model_prefix}")
            file_list_response = await manager_client.list_vfs_files(
                storage_name=registry_config.storage_name, directory=model_prefix
            )

            files = file_list_response.files
            if not files:
                log.warning(f"No files found for model: {model_prefix}")
                return 0

            log.info(f"Found {len(files)} files to download for model: {model_prefix}")

            # Initialize artifact download tracking in Redis with all file information
            revision = context.model.resolve_revision(ArtifactRegistryType.RESERVOIR)
            file_info_list = [
                (file_info.path, file_info.size or 0)
                for file_info in files
                if file_info.type != "directory"
            ]
            await self._redis_client.init_artifact_download(
                model_id=context.model.model_id,
                revision=revision,
                file_info_list=file_info_list,
            )

            # Create file downloader
            downloader = ReservoirVFSFileDownloader(
                client=manager_client,
                storage_name=registry_config.storage_name,
                redis_client=self._redis_client,
                model_id=context.model.model_id,
                revision=revision,
            )

            # Download each file individually
            total_bytes = 0
            for file_info in files:
                # Skip directories (they will be created automatically when files are downloaded)
                if file_info.type == "directory":
                    continue

                remote_file_path = file_info.path
                # Convert remote path to local path relative to dest_path
                # Remove model_prefix from the beginning to get relative path within model
                if remote_file_path.startswith(model_prefix):
                    relative_path = remote_file_path[len(model_prefix) :].lstrip("/")
                else:
                    relative_path = remote_file_path

                local_file_path = dest_path / relative_path
                bytes_downloaded = await downloader.download_file(
                    remote_file_path, local_file_path, file_info.size or 0
                )
                total_bytes += bytes_downloaded
                log.debug(f"Downloaded: {remote_file_path} ({bytes_downloaded} bytes)")

            log.info(f"VFS download completed: {model_prefix} -> {dest_path} ({total_bytes} bytes)")
            return total_bytes

        except Exception as e:
            log.error(f"VFS download failed for {model_prefix}: {str(e)}")
            raise

    async def _handle_object_storage_download(
        self,
        registry_config: ReservoirConfig,
        download_storage_name: str,
        context: ImportStepContext,
        model_prefix: str,
    ) -> tuple[list[tuple[FileObjectData, str]], int]:
        """Handle file downloads for object storage type."""
        # Use existing object storage download logic
        options = BucketCopyOptions(
            concurrency=4,
            progress_log_interval_bytes=8 * 1024 * 1024,  # 8MB intervals
        )

        # Initialize artifact download tracking
        revision = context.model.resolve_revision(ArtifactRegistryType.RESERVOIR)

        downloaded_files, bytes_copied = await self._stream_bucket_to_bucket(
            source_cfg=registry_config,
            storage_name=download_storage_name,
            storage_pool=context.storage_pool,
            options=options,
            model_id=context.model.model_id,
            revision=revision,
            progress_reporter=None,
            key_prefix=model_prefix,
        )

        return downloaded_files, bytes_copied

    def _get_s3_client(
        self, storage_pool: AbstractStoragePool, storage_name: str
    ) -> tuple[S3Client, str]:
        """Get S3 client for the specified storage"""
        # Get storage from pool and verify it's ObjectStorage type
        try:
            storage = storage_pool.get_storage(storage_name)
        except KeyError:
            raise StorageNotFoundError(f"Storage '{storage_name}' not found in pool")

        if not isinstance(storage, ObjectStorage):
            raise StorageNotFoundError(
                f"Storage '{storage_name}' is not an ObjectStorage type. "
                f"Reservoir import requires ObjectStorage for S3 operations."
            )

        # Use the configured bucket from ObjectStorage
        bucket_name = storage._bucket
        if not bucket_name:
            raise ObjectStorageBucketNotFoundError(
                f"No bucket configured for storage '{storage_name}'"
            )

        # Create S3Client from ObjectStorage configuration
        s3_client = S3Client(
            bucket_name=bucket_name,
            endpoint_url=storage._endpoint,
            region_name=storage._region,
            aws_access_key_id=storage._access_key,
            aws_secret_access_key=storage._secret_key,
        )
        return s3_client, bucket_name

    async def _stream_bucket_to_bucket(
        self,
        source_cfg: ReservoirConfig,
        storage_name: str,
        storage_pool: AbstractStoragePool,
        options: BucketCopyOptions,
        model_id: str,
        revision: str,
        progress_reporter: Optional[ProgressReporter],
        key_prefix: Optional[str] = None,
    ) -> tuple[list[tuple[FileObjectData, str]], int]:
        """Direct copy from Reservoir S3 to target storage"""
        dst_client, bucket_name = self._get_s3_client(storage_pool, storage_name)

        # Get storage from pool to access configuration
        try:
            storage = storage_pool.get_storage(storage_name)
        except KeyError:
            raise StorageNotFoundError(f"Storage '{storage_name}' not found in pool")

        if not isinstance(storage, ObjectStorage):
            raise StorageNotFoundError(f"Storage '{storage_name}' is not an ObjectStorage type")

        # List all objects under prefix
        src_s3_client = S3Client(
            bucket_name=bucket_name,
            endpoint_url=source_cfg.endpoint,
            region_name=source_cfg.object_storage_region,
            aws_access_key_id=source_cfg.object_storage_access_key,
            aws_secret_access_key=source_cfg.object_storage_secret_key,
        )

        target_keys, size_map, total_bytes = await self._list_all_keys_and_sizes(
            s3_client=src_s3_client,
            prefix=key_prefix,
        )

        if not target_keys:
            raise ArtifactStorageEmptyError()

        # Initialize artifact download tracking in Redis with all file information
        file_info_list = [(key, size_map.get(key, 0)) for key in target_keys]
        await self._redis_client.init_artifact_download(
            model_id=model_id,
            revision=revision,
            file_info_list=file_info_list,
        )

        log.trace(
            "[stream_bucket_to_bucket] start src_endpoint={} src_bucket={} src_prefix={} "
            "dst_storage={} dst_bucket={} objects={} total_bytes={} concurrency={}",
            source_cfg.endpoint,
            bucket_name,
            key_prefix,
            storage_name,
            bucket_name,
            len(target_keys),
            total_bytes,
            options.concurrency,
        )

        copied = 0
        sem = asyncio.Semaphore(options.concurrency)

        async def _copy_single_object(key: str) -> int:
            """
            Returns:
                The number of bytes copied.
            """
            async with sem:
                size = size_map.get(key, -1)
                log.trace("[stream_bucket_to_bucket] begin key={} size={}", key, size)

                download_chunk_size = storage._reservoir_download_chunk_size

                # Content-Type
                object_meta = await src_s3_client.get_object_meta(key)
                ctype = (
                    (object_meta.content_type if object_meta else None)
                    or mimetypes.guess_type(key)[0]
                    or "application/octet-stream"
                )

                data_stream = ReservoirS3FileDownloadStreamReader(
                    src_s3_client=src_s3_client,
                    key=key,
                    size=size,
                    options=options,
                    download_chunk_size=download_chunk_size,
                    content_type=ctype,
                    redis_client=self._redis_client,
                    model_id=model_id,
                    revision=revision,
                )

                part_size = storage._upload_chunk_size
                await dst_client.upload_stream(
                    data_stream,
                    key,
                    part_size=part_size,
                )

                log.trace("[stream_bucket_to_bucket] done key={} bytes={}", key, size)
                return size if size >= 0 else 0

        # TODO: Replace this with global semaphore
        sizes = await asyncio.gather(*(_copy_single_object(k) for k in target_keys))
        bytes_copied = sum(sizes)

        log.trace(
            "[stream_bucket_to_bucket] all done objects={} total_bytes={}", copied, total_bytes
        )

        # Build downloaded files list
        downloaded_files: list[tuple[FileObjectData, str]] = []
        for key in target_keys:
            size = size_map.get(key, 0)
            file_obj = FileObjectData(
                path=key,
                size=size,
                type="file",
                download_url="",  # Not applicable for reservoir
            )
            downloaded_files.append((file_obj, key))

        return downloaded_files, bytes_copied

    async def _list_all_keys_and_sizes(
        self,
        *,
        s3_client: S3Client,
        prefix: Optional[str] = None,
    ) -> tuple[list[str], dict[str, int], int]:
        """
        List all non-marker object keys in the given bucket (optionally under a prefix).

        Returns:
            keys: list of object keys
            size_map: mapping from key -> object size
            total_bytes: total sum of all object sizes
        """
        keys: list[str] = []
        size_map: dict[str, int] = {}
        total = 0

        log.trace(
            "[list] start bucket={} prefix={} endpoint={}",
            s3_client.bucket_name,
            prefix,
            s3_client.endpoint_url,
        )

        async with s3_client.session.client(
            "s3",
            endpoint_url=s3_client.endpoint_url,
            region_name=s3_client.region_name,
            aws_access_key_id=s3_client.aws_access_key_id,
            aws_secret_access_key=s3_client.aws_secret_access_key,
        ) as s3:
            paginator = s3.get_paginator("list_objects_v2")
            kwargs = {"Bucket": s3_client.bucket_name}
            if prefix:
                kwargs["Prefix"] = prefix

            async for page in paginator.paginate(**kwargs):
                for obj in page.get("Contents", []) or []:
                    key = obj["Key"]
                    if key.endswith("/"):  # directory marker skip
                        continue
                    size = int(obj.get("Size", 0))
                    keys.append(key)
                    size_map[key] = size
                    total += size

        log.trace("[list] done keys={} total_bytes={}", len(keys), total)
        return keys, size_map, total

    @override
    async def cleanup_stage(self, context: ImportStepContext) -> None:
        """Clean up files from download storage after successful transfer"""
        revision = context.model.resolve_revision(ArtifactRegistryType.RESERVOIR)
        model_prefix = f"{context.model.model_id}/{revision}"

        try:
            await self._download_storage.delete_file(model_prefix)
            log.info(f"[cleanup] Removed download files from reservoir storage: {model_prefix}")
        except Exception as e:
            log.warning(
                f"[cleanup] Failed to cleanup reservoir download storage: {model_prefix}: {str(e)}"
            )


class ReservoirVerifyStep(ModelVerifyStep):
    """Step to verify downloaded files in Reservoir model import"""

    @property
    @override
    def registry_type(self) -> ArtifactRegistryType:
        return ArtifactRegistryType.RESERVOIR


class ReservoirArchiveStep(ModelArchiveStep):
    """Step to move downloaded files to archive storage"""

    @property
    @override
    def registry_type(self) -> ArtifactRegistryType:
        return ArtifactRegistryType.RESERVOIR


# Utilities


def create_reservoir_import_pipeline(
    storage_pool: StoragePool,
    registry_configs: dict[str, Any],
    storage_step_mappings: dict[ArtifactStorageImportStep, str],
    transfer_manager: StorageTransferManager,
    artifact_verifier_ctx: ArtifactVerifierContext,
    event_producer: EventProducer,
    manager_http_clients: MutableMapping[str, ManagerHTTPClient],
    reservoir_client_config: ReservoirClientConfig,
    redis_client: ValkeyArtifactDownloadTrackingClient,
) -> ImportPipeline:
    """Create ImportPipeline for Reservoir based on storage step mappings."""
    steps: list[ImportStep[Any]] = []

    # Add steps based on what's present in storage_step_mappings
    if ArtifactStorageImportStep.DOWNLOAD in storage_step_mappings:
        # Get the download storage object from the pool
        download_storage_name = storage_step_mappings.get(ArtifactStorageImportStep.DOWNLOAD)
        if not download_storage_name:
            raise StorageStepRequiredStepNotProvided("Download storage not specified in mappings")

        download_storage = storage_pool.get_storage(download_storage_name)
        steps.append(
            ReservoirDownloadStep(
                registry_configs,
                download_storage,
                manager_http_clients,
                reservoir_client_config,
                redis_client,
            )
        )

    if ArtifactStorageImportStep.VERIFY in storage_step_mappings:
        steps.append(ReservoirVerifyStep(artifact_verifier_ctx, transfer_manager, event_producer))

    if ArtifactStorageImportStep.ARCHIVE in storage_step_mappings:
        steps.append(ReservoirArchiveStep(transfer_manager))

    return ImportPipeline(steps)


class TarExtractor:
    """Simple tar extractor that downloads and extracts VFS tar files."""

    _stream_reader: StreamReader

    def __init__(self, stream_reader: StreamReader):
        self._stream_reader = stream_reader

    async def extract_to(self, target_dir: Path) -> int:
        """
        Download tar file to temp location and extract to target directory.

        Args:
            target_dir: Directory where to extract the tar contents

        Returns:
            int: Total size of downloaded file in bytes
        """
        temp_file = None
        bytes_downloaded = 0

        try:
            # Create temporary file for tar
            with tempfile.NamedTemporaryFile(delete=False, suffix=".tar") as tf:
                temp_file = Path(tf.name)
                log.debug(f"Downloading artifact tar to temp directory: {temp_file}")

                # Download to temp file
                async with aiofiles.open(temp_file, "wb") as f:
                    async for chunk in self._stream_reader.read():
                        await f.write(chunk)
                        bytes_downloaded += len(chunk)

                log.debug(f"Downloaded {bytes_downloaded} bytes to {temp_file}")

            # Ensure target directory exists
            target_dir.mkdir(parents=True, exist_ok=True)

            # Extract tar file in executor to avoid blocking
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(None, self._extract_tar, temp_file, target_dir)

            log.info(f"Successfully extracted artifact tar to: {target_dir}")
            return bytes_downloaded

        finally:
            # Clean up temp file
            if temp_file and temp_file.exists():
                try:
                    temp_file.unlink()
                    log.debug(f"Cleaned up temp file: {temp_file}")
                except Exception as e:
                    log.warning(f"Failed to remove temp file {temp_file}: {e}")

    def _extract_tar(self, tar_path: Path, target_dir: Path) -> None:
        """Extract tar archive to target directory."""
        with tarfile.open(tar_path, "r") as tar:
            tar.extractall(path=target_dir)
        log.debug(f"Tar extraction completed: {tar_path} -> {target_dir}")



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import cast, override

from ai.backend.common.data.artifact.types import (
    ArtifactRegistryType,
    VerificationStepResult,
    VerifierResult,
)
from ai.backend.common.data.storage.types import ArtifactStorageImportStep
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.events.event_types.artifact.anycast import (
    ModelVerifyDoneEvent,
    ModelVerifyingEvent,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.storage.context_types import ArtifactVerifierContext
from ai.backend.storage.exception import (
    ArtifactVerificationFailedError,
    ArtifactVerifyStorageTypeInvalid,
    NotImplementedAPI,
    StorageStepRequiredStepNotProvided,
)
from ai.backend.storage.services.artifacts.storage_transfer import StorageTransferManager
from ai.backend.storage.services.artifacts.types import (
    DownloadStepResult,
    ImportStep,
    ImportStepContext,
    VerifyStepResult,
)
from ai.backend.storage.storages.vfs_storage import VFSStorage

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ModelVerifyStep(ImportStep[DownloadStepResult], ABC):
    """Base class for model verification steps that transfer files and run artifact verification"""

    _artifact_verifier_ctx: ArtifactVerifierContext
    _transfer_manager: StorageTransferManager
    _event_producer: EventProducer

    def __init__(
        self,
        artifact_verifier_ctx: ArtifactVerifierContext,
        transfer_manager: StorageTransferManager,
        event_producer: EventProducer,
    ) -> None:
        self._artifact_verifier_ctx = artifact_verifier_ctx
        self._transfer_manager = transfer_manager
        self._event_producer = event_producer

    @property
    def step_type(self) -> ArtifactStorageImportStep:
        return ArtifactStorageImportStep.VERIFY

    @property
    @abstractmethod
    def registry_type(self) -> ArtifactRegistryType:
        """Registry type used for revision resolution"""
        raise NotImplementedAPI

    @override
    async def execute(
        self, context: ImportStepContext, input_data: DownloadStepResult
    ) -> VerifyStepResult:
        source_storage_name = context.storage_step_mappings.get(ArtifactStorageImportStep.DOWNLOAD)
        dst_storage_name = context.storage_step_mappings.get(ArtifactStorageImportStep.VERIFY)
        if source_storage_name is None:
            raise StorageStepRequiredStepNotProvided(
                "No storage mapping provided for DOWNLOAD step"
            )
        if dst_storage_name is None:
            raise StorageStepRequiredStepNotProvided("No storage mapping provided for VERIFY step")

        revision = context.model.resolve_revision(self.registry_type)
        model_prefix = f"{context.model.model_id}/{revision}"

        # Send model verifying event
        await self._event_producer.anycast_event(
            ModelVerifyingEvent(
                model_id=context.model.model_id,
                revision=revision,
                registry_type=self.registry_type,
                registry_name=context.registry_name,
            )
        )

        await self._transfer_manager.transfer_directory(
            source_storage_name=source_storage_name,
            dest_storage_name=dst_storage_name,
            source_prefix=model_prefix,
            dest_prefix=model_prefix,
        )

        dst_storage = context.storage_pool.get_storage(dst_storage_name)
        if not isinstance(dst_storage, VFSStorage):
            raise ArtifactVerifyStorageTypeInvalid("Verify step requires VFS storage type")
        dst_storage = cast(VFSStorage, dst_storage)

        # Collect verification results from all verifiers
        verifier_results: dict[str, VerifierResult] = {}
        verification_success = True

        for verifier_name, verifier in self._artifact_verifier_ctx._verifiers.items():
            dst_path = dst_storage.resolve_path(model_prefix)
            verifier_start_time = datetime.now(timezone.utc)
            log.info(
                f"Starting artifact verification using '{verifier_name}', dst_path: {dst_path}"
            )
            try:
                result = await verifier.verify(dst_path, context)
                verifier_end_time = datetime.now(timezone.utc)
                elapsed_time = (verifier_end_time - verifier_start_time).total_seconds()

                # Create VerifierResult object
                verifier_results[verifier_name] = VerifierResult(
                    success=result.infected_count == 0,
                    infected_count=result.infected_count,
                    scanned_at=verifier_start_time,
                    scan_time=elapsed_time,
                    scanned_count=result.scanned_count,
                    metadata=result.metadata,
                )

                if result.infected_count > 0:
                    verification_success = False
                    log.warning(
                        f"Artifact verification using '{verifier_name}' found {result.infected_count} infected files"
                    )
                else:
                    log.info(
                        f"Artifact verification using '{verifier_name}' completed successfully"
                    )

            except Exception as e:
                verification_success = False
                verifier_end_time = datetime.now(timezone.utc)
                elapsed_time = (verifier_end_time - verifier_start_time).total_seconds()
                verifier_results[verifier_name] = VerifierResult(
                    success=False,
                    infected_count=0,
                    scanned_at=verifier_start_time,
                    scan_time=elapsed_time,
                    scanned_count=0,
                    metadata={},
                    error=str(e),
                )
                log.error(f"Artifact verification using '{verifier_name}' failed: {e}")

        # Create complete verification result
        verification_result = VerificationStepResult(
            verifiers=verifier_results,
        )

        # Send ModelVerifyDoneEvent with collected results
        await self._event_producer.anycast_event(
            ModelVerifyDoneEvent(
                model_id=context.model.model_id,
                revision=revision,
                registry_type=self.registry_type,
                registry_name=context.registry_name,
                success=verification_success,
                verification_result=verification_result.model_dump(),
            )
        )

        # Raise error if any verification failed
        if not verification_success:
            raise ArtifactVerificationFailedError(
                f"Artifact '{model_prefix}' verification failed. See verification_result for details."
            )

        return VerifyStepResult(
            verified_files=input_data.downloaded_files,
            storage_name=dst_storage_name,
            total_bytes=input_data.total_bytes,
        )

    @override
    async def cleanup_stage(self, context: ImportStepContext) -> None:
        """Clean up files from verify storage on failure"""
        verify_storage = context.storage_step_mappings.get(ArtifactStorageImportStep.VERIFY)
        if not verify_storage:
            return

        # Delete entire model (cleaning up individual files is complex)
        revision = context.model.resolve_revision(self.registry_type)
        model_prefix = f"{context.model.model_id}/{revision}"

        try:
            storage = context.storage_pool.get_storage(verify_storage)
            await storage.delete_file(model_prefix)
            log.info(f"[cleanup] Removed verify files: {verify_storage}:{model_prefix}")
        except Exception as e:
            log.warning(
                f"[cleanup] Failed to cleanup verify: {verify_storage}:{model_prefix}: {str(e)}"
            )


class ModelArchiveStep(ImportStep[VerifyStepResult], ABC):
    """Base class for model archiving steps that transfer files to archive storage"""

    def __init__(self, transfer_manager: StorageTransferManager) -> None:
        self._transfer_manager = transfer_manager

    @property
    def step_type(self) -> ArtifactStorageImportStep:
        return ArtifactStorageImportStep.ARCHIVE

    @property
    @abstractmethod
    def registry_type(self) -> ArtifactRegistryType:
        """Registry type used for revision resolution"""
        raise NotImplementedAPI

    @override
    async def execute(
        self,
        context: ImportStepContext,
        input_data: VerifyStepResult,
    ) -> None:
        download_storage = input_data.storage_name
        archive_storage = context.storage_step_mappings.get(ArtifactStorageImportStep.ARCHIVE)

        if not archive_storage:
            raise StorageStepRequiredStepNotProvided("No storage mapping provided for ARCHIVE step")

        # No need to move if download and archive storage are the same
        if download_storage == archive_storage:
            log.info(
                f"Archive step skipped - download and archive storage are the same: {archive_storage}"
            )
            return

        log.info(f"Starting archive transfer: {download_storage} -> {archive_storage}")

        archieved_file_cnt = 0
        # Move each file from download storage to archive storage
        for file_info, storage_key in input_data.verified_files:
            try:
                archieved_file_cnt += await self._transfer_manager.transfer_directory(
                    source_storage_name=download_storage,
                    dest_storage_name=archive_storage,
                    source_prefix=storage_key,
                    dest_prefix=storage_key,
                )
                log.debug(f"Transferred file to archive: {storage_key}")
            except Exception as e:
                log.error(f"Failed to transfer file to archive: {storage_key}: {str(e)}")
                raise

        log.info(
            f"Archive transfer completed: {download_storage} -> {archive_storage}, "
            f"files={archieved_file_cnt}"
        )

    @override
    async def cleanup_stage(self, context: ImportStepContext) -> None:
        """Clean up files from archive storage on failure"""
        archive_storage = context.storage_step_mappings.get(ArtifactStorageImportStep.ARCHIVE)
        if not archive_storage:
            return

        # Delete entire model (cleaning up individual files is complex)
        revision = context.model.resolve_revision(self.registry_type)
        model_prefix = f"{context.model.model_id}/{revision}"

        try:
            storage = context.storage_pool.get_storage(archive_storage)
            await storage.delete_file(model_prefix)
            log.info(f"[cleanup] Removed archive files: {archive_storage}:{model_prefix}")
        except Exception as e:
            log.warning(
                f"[cleanup] Failed to cleanup archive: {archive_storage}:{model_prefix}: {str(e)}"
            )



import logging

from ai.backend.common.dto.storage.response import (
    VFSFileInfo,
    VFSFileMetaResponse,
    VFSListFilesResponse,
)
from ai.backend.common.types import StreamReader
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.storage.exception import StorageNotFoundError, StorageTypeInvalidError
from ai.backend.storage.storages.storage_pool import StoragePool
from ai.backend.storage.storages.vfs_storage import VFSStorage

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class VFSStorageService:
    """
    Service class for VFS storage operations.
    Provides a service layer similar to ObjectStorageService but for VFS storage.
    """

    _storage_pool: StoragePool

    def __init__(self, storage_pool: StoragePool) -> None:
        self._storage_pool = storage_pool

    async def stream_upload(
        self,
        storage_name: str,
        filepath: str,
        data_stream: StreamReader,
    ) -> None:
        """
        Upload a file to VFS using streaming.

        Args:
            storage_name: Name of the VFS storage configuration
            filepath: Path to the file to upload
            data_stream: Async iterator of file data chunks
        """
        storage = self._resolve_storage(storage_name)
        await storage.stream_upload(filepath, data_stream)

    async def stream_download(self, storage_name: str, filepath: str) -> StreamReader:
        """
        Download a file from VFS using streaming.

        Args:
            storage_name: Name of the VFS storage configuration
            filepath: Path to the file to download

        Returns:
            FileStream: Stream for reading file data
        """
        storage = self._resolve_storage(storage_name)
        return await storage.stream_download(filepath)

    async def get_file_meta(self, storage_name: str, filepath: str) -> VFSFileMetaResponse:
        """
        Get file information.

        Args:
            storage_name: Name of the VFS storage configuration
            filepath: Path to the file

        Returns:
            VFSFileMetaResponse with file metadata
        """
        storage = self._resolve_storage(storage_name)
        return await storage.get_file_info(filepath)

    async def delete_file(self, storage_name: str, filepath: str) -> None:
        """
        Delete a file or directory.

        Args:
            storage_name: Name of the VFS storage configuration
            filepath: Path to the file/directory to delete
        """
        storage = self._resolve_storage(storage_name)
        await storage.delete_file(filepath)

    async def list_files(self, storage_name: str, directory: str) -> list[VFSFileInfo]:
        """
        List files and directories in a directory.

        Args:
            storage_name: Name of the VFS storage configuration
            directory: Directory path to list

        Returns:
            List of file/directory information
        """
        storage = self._resolve_storage(storage_name)
        return await storage.list_directory(directory)

    async def list_files_recursive(self, storage_name: str, directory: str) -> VFSListFilesResponse:
        """
        Recursively list all files in a directory and its subdirectories.

        Args:
            storage_name: Name of the VFS storage configuration
            directory: Directory path to start listing from

        Returns:
            VFSListFilesResponse containing all files found recursively
        """
        storage = self._resolve_storage(storage_name)
        return await storage.list_files_recursive(directory)

    async def create_directory(self, storage_name: str, directory: str) -> None:
        """
        Create a directory.

        Args:
            storage_name: Name of the VFS storage configuration
            directory: Directory path to create
        """
        storage = self._resolve_storage(storage_name)
        await storage.create_directory(directory)

    async def get_disk_usage(self, storage_name: str) -> dict:
        """
        Get disk usage information for the storage.

        Args:
            storage_name: Name of the VFS storage configuration

        Returns:
            Dictionary with capacity and used bytes information
        """
        storage = self._resolve_storage(storage_name)
        return await storage.get_disk_usage()

    def _resolve_storage(self, storage_name: str) -> VFSStorage:
        """
        Resolve VFS storage by name from storage pool.

        Args:
            storage_name: Name of the VFS storage configuration

        Returns:
            VFSStorage instance

        Raises:
            StorageNotFoundError: If storage is not found or not a VFS storage
        """
        try:
            storage = self._storage_pool.get_storage(storage_name)
            if not isinstance(storage, VFSStorage):
                raise StorageTypeInvalidError(f"Storage '{storage_name}' is not a VFS storage")
            return storage
        except KeyError:
            raise StorageNotFoundError(f"No VFS storage configuration found for: {storage_name}")



from http import HTTPStatus
from typing import Optional, Protocol

from ai.backend.common.api_handlers import APIResponse, BodyParam, PathParam, api_handler
from ai.backend.common.dto.storage.path import QuotaScopeKeyPath, VFolderKeyPath, VolumeIDPath
from ai.backend.common.dto.storage.request import (
    CloneVFolderReq,
    GetVFolderMetaReq,
    QuotaScopeReq,
)
from ai.backend.common.dto.storage.response import (
    GetVolumeResponse,
    GetVolumesResponse,
    VFolderMetadataResponse,
)
from ai.backend.common.types import QuotaConfig, VFolderID, VolumeID

from ...volumes.types import (
    QuotaScopeKey,
    QuotaScopeMeta,
    VFolderKey,
    VFolderMeta,
    VolumeMeta,
)


class VFolderServiceProtocol(Protocol):
    async def get_volume(self, volume_id: VolumeID) -> VolumeMeta: ...

    async def get_volumes(self) -> list[VolumeMeta]: ...

    async def create_quota_scope(
        self, quota_scope_key: QuotaScopeKey, options: Optional[QuotaConfig]
    ) -> None: ...

    async def get_quota_scope(self, quota_scope_key: QuotaScopeKey) -> QuotaScopeMeta: ...

    async def update_quota_scope(
        self, quota_scope_key: QuotaScopeKey, options: Optional[QuotaConfig]
    ) -> None: ...

    async def delete_quota_scope(self, quota_scope_key: QuotaScopeKey) -> None: ...

    async def create_vfolder(self, vfolder_key: VFolderKey) -> None: ...

    async def clone_vfolder(self, vfolder_key: VFolderKey, dst_vfolder_id: VFolderID) -> None: ...

    async def get_vfolder_info(self, vfolder_key: VFolderKey, subpath: str) -> VFolderMeta: ...

    async def delete_vfolder(self, vfolder_key: VFolderKey) -> None: ...


class VFolderHandler:
    _storage_service: VFolderServiceProtocol

    def __init__(self, storage_service: VFolderServiceProtocol) -> None:
        self._storage_service = storage_service

    @api_handler
    async def get_volume(self, path: PathParam[VolumeIDPath]) -> APIResponse:
        volume_meta = await self._storage_service.get_volume(path.parsed.volume_id)
        return APIResponse.build(
            status_code=HTTPStatus.OK,
            response_model=GetVolumeResponse(
                item=volume_meta.to_field(),
            ),
        )

    @api_handler
    async def get_volumes(self) -> APIResponse:
        volume_meta_list = await self._storage_service.get_volumes()
        return APIResponse.build(
            status_code=HTTPStatus.OK,
            response_model=GetVolumesResponse(
                items=[volume.to_field() for volume in volume_meta_list],
            ),
        )

    @api_handler
    async def create_quota_scope(
        self, path: PathParam[QuotaScopeKeyPath], body: BodyParam[QuotaScopeReq]
    ) -> APIResponse:
        quota_scope_key = QuotaScopeKey.from_quota_scope_path(path.parsed)
        await self._storage_service.create_quota_scope(quota_scope_key, body.parsed.options)
        return APIResponse.no_content(status_code=HTTPStatus.NO_CONTENT)

    @api_handler
    async def get_quota_scope(self, path: PathParam[QuotaScopeKeyPath]) -> APIResponse:
        quota_scope_key = QuotaScopeKey.from_quota_scope_path(path.parsed)
        quota_scope = await self._storage_service.get_quota_scope(quota_scope_key)
        return APIResponse.build(
            status_code=HTTPStatus.OK,
            response_model=quota_scope.to_response(),
        )

    @api_handler
    async def update_quota_scope(
        self, path: PathParam[QuotaScopeKeyPath], body: BodyParam[QuotaScopeReq]
    ) -> APIResponse:
        quota_scope_key = QuotaScopeKey.from_quota_scope_path(path.parsed)
        await self._storage_service.update_quota_scope(quota_scope_key, body.parsed.options)
        return APIResponse.no_content(status_code=HTTPStatus.NO_CONTENT)

    @api_handler
    async def delete_quota_scope(self, path: PathParam[QuotaScopeKeyPath]) -> APIResponse:
        quota_scope_key = QuotaScopeKey.from_quota_scope_path(path.parsed)
        await self._storage_service.delete_quota_scope(quota_scope_key)
        return APIResponse.no_content(status_code=HTTPStatus.NO_CONTENT)

    @api_handler
    async def create_vfolder(self, path: PathParam[VFolderKeyPath]) -> APIResponse:
        vfolder_key = VFolderKey.from_vfolder_path(path.parsed)
        await self._storage_service.create_vfolder(vfolder_key)
        return APIResponse.no_content(status_code=HTTPStatus.NO_CONTENT)

    @api_handler
    async def clone_vfolder(
        self, path: PathParam[VFolderKeyPath], body: BodyParam[CloneVFolderReq]
    ) -> APIResponse:
        vfolder_key = VFolderKey.from_vfolder_path(path.parsed)
        await self._storage_service.clone_vfolder(vfolder_key, body.parsed.dst_vfolder_id)
        return APIResponse.no_content(status_code=HTTPStatus.NO_CONTENT)

    @api_handler
    async def get_vfolder_info(
        self, path: PathParam[VFolderKeyPath], body: BodyParam[GetVFolderMetaReq]
    ) -> APIResponse:
        vfolder_key = VFolderKey.from_vfolder_path(path.parsed)
        vfolder_meta = await self._storage_service.get_vfolder_info(
            vfolder_key, body.parsed.subpath
        )
        return APIResponse.build(
            status_code=HTTPStatus.OK,
            response_model=VFolderMetadataResponse(
                item=vfolder_meta.to_field(),
            ),
        )

    @api_handler
    async def delete_vfolder(self, path: PathParam[VFolderKeyPath]) -> APIResponse:
        vfolder_key = VFolderKey.from_vfolder_path(path.parsed)
        await self._storage_service.delete_vfolder(vfolder_key)
        return APIResponse.no_content(status_code=HTTPStatus.NO_CONTENT)



import mimetypes
from typing import override

from ai.backend.common.artifact_storage import AbstractStorage
from ai.backend.common.dto.storage.response import (
    ObjectMetaResponse,
    PresignedDownloadObjectResponse,
    PresignedUploadObjectResponse,
)
from ai.backend.common.types import StreamReader
from ai.backend.storage.client.s3 import S3Client
from ai.backend.storage.config.unified import (
    ObjectStorageConfig,
    PresignedDownloadConfig,
    PresignedUploadConfig,
)
from ai.backend.storage.exception import (
    FileStreamDownloadError,
    FileStreamUploadError,
    ObjectInfoFetchError,
    ObjectStorageObjectDeletionError,
    PresignedDownloadURLGenerationError,
    PresignedUploadURLGenerationError,
    StorageBucketFileNotFoundError,
)


class ObjectStorage(AbstractStorage):
    _name: str
    _endpoint: str
    _access_key: str
    _secret_key: str
    _bucket: str
    _region: str
    _upload_chunk_size: int
    _download_chunk_size: int
    _reservoir_download_chunk_size: int
    _presigned_upload_config: PresignedUploadConfig
    _presigned_download_config: PresignedDownloadConfig

    def __init__(self, name: str, cfg: ObjectStorageConfig) -> None:
        self._name = name
        self._endpoint = cfg.endpoint
        self._access_key = cfg.access_key
        self._secret_key = cfg.secret_key
        self._bucket = cfg.buckets[0]
        self._region = cfg.region
        self._upload_chunk_size = cfg.upload_chunk_size
        self._download_chunk_size = cfg.download_chunk_size
        self._reservoir_download_chunk_size = cfg.reservoir_download_chunk_size
        self._presigned_upload_config = cfg.presigned_upload
        self._presigned_download_config = cfg.presigned_download

    @override
    async def stream_upload(
        self,
        filepath: str,
        data_stream: StreamReader,
    ) -> None:
        """
        Upload a file to S3 using streaming.

        Args:
            filepath: Path to the file to upload
            data_stream: Async iterator of file data chunks
            content_type: Content type of the file
        """
        try:
            part_size = self._upload_chunk_size
            s3_client = self._get_s3_client()
            await s3_client.upload_stream(
                data_stream,
                filepath,
                part_size=part_size,
            )
        except Exception as e:
            raise FileStreamUploadError("Upload failed") from e

    @override
    async def stream_download(self, filepath: str) -> StreamReader:
        """
        Download a file from S3 using streaming.

        Args:
            filepath: Path to the file to download

        Returns:
            FileStream: Stream for reading file data
        """
        try:
            s3_client = self._get_s3_client()
            object_meta = await s3_client.get_object_meta(filepath)
            ctype = (
                (object_meta.content_type if object_meta else None)
                or mimetypes.guess_type(filepath)[0]
                or "application/octet-stream"
            )
            chunk_size = self._download_chunk_size
            return s3_client.download_stream(filepath, chunk_size=chunk_size, content_type=ctype)

        except Exception as e:
            raise FileStreamDownloadError("Download failed") from e

    @override
    async def get_file_info(self, filepath: str) -> ObjectMetaResponse:
        """
        Get object information.

        Args:
            filepath: Path to the file to download

        Returns:
            ObjectMetaResponse with object metadata
        """
        try:
            s3_client = self._get_s3_client()
            object_info = await s3_client.get_object_meta(filepath)

            if object_info is None:
                raise StorageBucketFileNotFoundError()

            return ObjectMetaResponse(
                content_length=object_info.content_length,
                content_type=object_info.content_type,
                last_modified=object_info.last_modified,
                etag=object_info.etag,
                metadata=object_info.metadata,
            )

        except Exception as e:
            raise ObjectInfoFetchError(f"Get object info failed: {str(e)}") from e

    @override
    async def delete_file(self, filepath: str) -> None:
        """
        Delete an object and all its contents.

        Args:
            filepath: Path of the file to delete
        """
        try:
            s3_client = self._get_s3_client()
            await s3_client.delete_object(filepath)
        except Exception as e:
            raise ObjectStorageObjectDeletionError(f"Delete object failed: {str(e)}") from e

    async def generate_presigned_upload_url(self, key: str) -> PresignedUploadObjectResponse:
        """
        Generate presigned upload URL.

        Args:
            key: Path to the file to upload

        Returns:
            PresignedUploadObjectResponse with URL and fields
        """
        try:
            s3_client = self._get_s3_client()
            presigned_upload_config = self._presigned_upload_config

            presigned_data = await s3_client.generate_presigned_upload_url(
                key,
                expiration=presigned_upload_config.expiration,
                content_length_range=(
                    presigned_upload_config.min_size,
                    presigned_upload_config.max_size,
                )
                if presigned_upload_config.min_size and presigned_upload_config.max_size
                else None,
            )

            if presigned_data is None:
                raise PresignedUploadURLGenerationError()

            return PresignedUploadObjectResponse(
                url=presigned_data.url, fields=presigned_data.fields
            )

        except Exception as e:
            raise PresignedUploadURLGenerationError() from e

    async def generate_presigned_download_url(
        self,
        filepath: str,
    ) -> PresignedDownloadObjectResponse:
        """
        Generate presigned download URL.

        Args:
            filepath: Path to the file to download

        Returns:
            PresignedDownloadResponse with URL
        """
        try:
            s3_client = self._get_s3_client()

            presigned_url = await s3_client.generate_presigned_download_url(
                filepath,
                expiration=self._presigned_download_config.expiration,
            )

            if presigned_url is None:
                raise PresignedDownloadURLGenerationError()

            return PresignedDownloadObjectResponse(url=presigned_url)

        except Exception as e:
            raise PresignedDownloadURLGenerationError() from e

    async def list_objects_with_prefix(self, prefix: str) -> list[str]:
        """
        List all object keys in the storage with the given prefix.

        Args:
            prefix: The prefix to filter object keys

        Returns:
            list[str]: List of object keys matching the prefix
        """
        try:
            s3_client = self._get_s3_client()
            return await s3_client.list_objects_with_prefix(prefix)
        except Exception as e:
            raise ObjectInfoFetchError(f"List objects failed: {str(e)}") from e

    def _get_s3_client(self) -> S3Client:
        return S3Client(
            bucket_name=self._bucket,
            endpoint_url=self._endpoint,
            region_name=self._region,
            aws_access_key_id=self._access_key,
            aws_secret_access_key=self._secret_key,
        )



from __future__ import annotations

import enum
from collections.abc import AsyncIterator
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path, PurePath
from typing import Any, Final, Mapping, Optional, override

import attrs
import trafaret as t
from aiohttp import BodyPartReader, MultipartReader, web
from pydantic import BaseModel

from ai.backend.common import validators as tx
from ai.backend.common.types import QuotaConfig, StreamReader, VFolderID

__all__ = (
    "Sentinel",
    "SENTINEL",
    "FSPerfMetric",
    "CapacityUsage",
    "VolumeInfo",
    "VFolderID",
    "TreeUsage",
    "QuotaConfig",
    "Stat",
    "DirEntry",
    "DirEntryType",
    "HealthResponse",
)


class Sentinel(enum.Enum):
    token = 0


SENTINEL: Final = Sentinel.token


@attrs.define(slots=True, frozen=True)
class FSPerfMetric:
    # iops
    iops_read: int
    iops_write: int
    # thruput
    io_bytes_read: int
    io_bytes_write: int
    # latency
    io_usec_read: float
    io_usec_write: float


@attrs.define(slots=True, frozen=True)
class CapacityUsage:
    used_bytes: int
    capacity_bytes: int


@attrs.define(slots=True, frozen=True)
class QuotaUsage:
    used_bytes: int
    limit_bytes: int


@attrs.define(slots=True, frozen=True)
class VolumeInfo:
    backend: str
    path: Path
    fsprefix: Optional[PurePath]
    options: Optional[Mapping[str, Any]]

    @classmethod
    def as_trafaret(cls) -> t.Trafaret:
        return t.Dict(
            {
                t.Key("backend"): t.String,
                t.Key("path"): tx.Path(type="dir"),
                t.Key("fsprefix", default="."): tx.PurePath(relative_only=True),
                t.Key("options", default=None): t.Null | t.Mapping(t.String, t.Any),
            },
        )


@attrs.define(slots=True, frozen=True)
class TreeUsage:
    file_count: int  # TODO: refactor using DecimalSize
    used_bytes: int  # TODO: refactor using DecimalSize


@attrs.define(slots=True, frozen=True)
class Stat:
    size: int
    owner: str
    mode: int
    modified: datetime
    created: datetime


class DirEntryType(enum.Enum):
    FILE = 0
    DIRECTORY = 1
    SYMLINK = 2


@attrs.define(auto_attribs=True, slots=True, frozen=True)
class DirEntry:
    name: str
    path: Path
    type: DirEntryType
    stat: Stat
    symlink_target: str


@dataclass
class BucketCopyOptions:
    concurrency: int
    progress_log_interval_bytes: int


_DEFAULT_UPLOAD_FILE_CHUNKS = 8192  # Default chunk size for streaming uploads


class MultipartFileUploadStreamReader(StreamReader):
    _file_reader: MultipartReader

    def __init__(self, file_reader: MultipartReader, content_type: Optional[str]) -> None:
        self._file_reader = file_reader
        self._content_type = content_type

    @override
    async def read(self) -> AsyncIterator[bytes]:
        file_part = await self._file_reader.next()
        while file_part and not getattr(file_part, "filename", None):
            await file_part.release()
            file_part = await self._file_reader.next()

        # TODO: Make exception class
        if file_part is None:
            raise web.HTTPBadRequest(reason='No file part found (expected field "file")')
        if not isinstance(file_part, BodyPartReader):
            raise web.HTTPBadRequest(reason="Invalid file part")

        while True:
            chunk = await file_part.read_chunk(_DEFAULT_UPLOAD_FILE_CHUNKS)
            if not chunk:
                break
            yield chunk

    @override
    def content_type(self) -> Optional[str]:
        return self._content_type


class HealthResponse(BaseModel):
    """Standard health check response"""

    status: str
    version: str
    component: str



from __future__ import annotations

from typing import Awaitable, Callable, Mapping, TypeAlias

import aiohttp_cors
from aiohttp import web
from aiohttp.typedefs import Middleware

WebRequestHandler: TypeAlias = Callable[
    [web.Request],
    Awaitable[web.StreamResponse],
]
WebMiddleware: TypeAlias = Middleware

CORSOptions: TypeAlias = Mapping[str, aiohttp_cors.ResourceOptions]



import enum
import logging
from contextlib import asynccontextmanager as actxmgr
from datetime import datetime
from datetime import timezone as tz
from pathlib import PurePath
from typing import Any, AsyncIterator, Optional, Union

import trafaret as t
from aiohttp import web

from ai.backend.common.json import dump_json_str
from ai.backend.logging import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class CheckParamSource(enum.Enum):
    BODY = 0
    QUERY = 1


def fstime2datetime(t: Union[float, int]) -> datetime:
    return datetime.utcfromtimestamp(t).replace(tzinfo=tz.utc)


@actxmgr
async def check_params(
    request: web.Request,
    checker: Optional[t.Trafaret],
    *,
    read_from: CheckParamSource = CheckParamSource.BODY,
    auth_required: bool = True,
) -> AsyncIterator[Any]:
    if checker is None:
        if request.can_read_body:
            raise web.HTTPBadRequest(
                text=dump_json_str(
                    {
                        "type": "https://api.backend.ai/probs/storage/malformed-request",
                        "title": "Malformed request (request body should be empty)",
                    },
                ),
                content_type="application/problem+json",
            )
    else:
        if read_from == CheckParamSource.BODY:
            raw_params = await request.json()
        elif read_from == CheckParamSource.QUERY:
            raw_params = request.query
        else:
            raise ValueError("Invalid source for check_params() helper")
    try:
        if checker is None:
            yield None
        else:
            yield checker.check(raw_params)
    except t.DataError as e:
        log.debug("check_params IV error", exc_info=e)
        raise web.HTTPBadRequest(
            text=dump_json_str(
                {
                    "type": "https://api.backend.ai/probs/storage/invalid-api-params",
                    "title": "Invalid API parameters",
                    "data": e.as_dict(),
                },
            ),
            content_type="application/problem+json",
        )
    except NotImplementedError:
        raise web.HTTPBadRequest(
            text=dump_json_str(
                {
                    "type": "https://api.backend.ai/probs/storage/unsupported-operation",
                    "title": "Unsupported operation by the storage backend",
                },
            ),
            content_type="application/problem+json",
        )


async def log_manager_api_entry(
    log: Union[logging.Logger, BraceStyleAdapter],
    name: str,
    params: Any,
) -> None:
    if params is not None:
        if "src_vfid" in params and "dst_vfid" in params:
            log.info(
                "ManagerAPI::{}(v:{}, f:{} -> dst_v: {}, dst_f:{})",
                name.upper(),
                params["src_volume"],
                params["src_vfid"],
                params["dst_volume"],
                params["dst_vfid"],
            )
        elif "relpaths" in params:
            log.info(
                "ManagerAPI::{}(v:{}, f:{}, p*:{})",
                name.upper(),
                params["volume"],
                params["vfid"],
                str(params["relpaths"][0]) + "...",
            )
        elif "relpath" in params:
            log.info(
                "ManagerAPI::{}(v:{}, f:{}, p:{})",
                name.upper(),
                params["volume"],
                params["vfid"],
                params["relpath"],
            )
        elif "vfid" in params:
            log.info(
                "ManagerAPI::{}(v:{}, f:{})",
                name.upper(),
                params["volume"],
                params["vfid"],
            )
        elif "volume" in params:
            log.info(
                "ManagerAPI::{}(v:{})",
                name.upper(),
                params["volume"],
            )
        return
    log.info(
        "ManagerAPI::{}()",
        name.upper(),
    )


async def log_manager_api_entry_new(
    log: Union[logging.Logger, BraceStyleAdapter],
    name: str,
    params: Any,
) -> None:
    from .volumes.types import LoggingInternalMeta

    if params is None:
        log.info(
            "ManagerAPI::{}()",
            name.upper(),
        )
    elif isinstance(params, LoggingInternalMeta):
        log.info(
            "ManagerAPI::{}({})",
            name.upper(),
            params.to_logging_str(),
        )
    else:
        log.info(
            "ManagerAPI::{}({})",
            name.upper(),
            str(params),
        )


async def log_client_api_entry(
    log: Union[logging.Logger, BraceStyleAdapter],
    name: str,
    params: Any,
) -> None:
    from .volumes.types import LoggingInternalMeta

    if params is None:
        log.info(
            "ClientFacingAPI::{}()",
            name.upper(),
        )
    elif isinstance(params, LoggingInternalMeta):
        log.info(
            "ClientFacingAPI::{}({})",
            name.upper(),
            params.to_logging_str(),
        )
    else:
        log.info(
            "ClientFacingAPI::{}({})",
            name.upper(),
            str(params),
        )


def normalize_filepath(filepath: str) -> str:
    """
    Normalize a filepath by removing path traversal components and extra slashes.

    Args:
        filepath: The filepath to normalize

    Returns:
        Normalized filepath string

    Raises:
        ValueError: If the filepath contains invalid characters or path traversal
    """
    if not filepath:
        raise ValueError("Filepath cannot be empty")

    # Convert to PurePath to handle cross-platform path normalization
    path = PurePath(filepath)

    # Check for path traversal attempts
    for part in path.parts:
        if part in (".", ".."):
            raise ValueError(f"Path traversal not allowed: {filepath}")
        if "\x00" in part:  # Null byte
            raise ValueError(f"Invalid character in filepath: {filepath}")

    # Convert back to string with forward slashes (POSIX style)
    normalized = str(path).replace("\\", "/")

    # Remove leading slash if present (we want relative paths)
    if normalized.startswith("/"):
        normalized = normalized[1:]

    return normalized



from __future__ import annotations

import enum
from typing import Self


class StorageBgtaskName(enum.StrEnum):
    """Background task names for storage component."""

    CLONE_VFOLDER = "clone_vfolder"
    DELETE_VFOLDER = "delete_vfolder"

    @classmethod
    def from_str(cls, value: str) -> Self:
        """Create instance from string value."""
        return cls(value)



from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from pathlib import Path, PurePosixPath
from typing import (
    Any,
    AsyncIterator,
    ClassVar,
    Final,
    FrozenSet,
    Mapping,
    Optional,
    Sequence,
    final,
)

from ai.backend.common.defs import DEFAULT_VFOLDER_PERMISSION_MODE
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events.dispatcher import EventDispatcher, EventProducer
from ai.backend.common.types import BinarySize, HardwareMetadata, QuotaScopeID
from ai.backend.logging import BraceStyleAdapter

from ..exception import InvalidSubpathError, VFolderNotFoundError
from ..types import (
    CapacityUsage,
    DirEntry,
    FSPerfMetric,
    QuotaConfig,
    QuotaUsage,
    TreeUsage,
    VFolderID,
    VolumeInfo,
)
from ..watcher import WatcherClient

# Available capabilities of a volume implementation
CAP_VFOLDER: Final = "vfolder"  # ability to create vfolder
CAP_METRIC: Final = "metric"  # ability to report disk related metrics
CAP_QUOTA: Final = "quota"  # ability to manage quota limits
# ability to scan filesystem size fast (e.g. by API)
CAP_FAST_FS_SIZE: Final = "fast-fs-size"
# ability to scan number of files in vFolder fast (e.g. by API)
CAP_FAST_SCAN: Final = "fast-scan"
# ability to scan vFolder size fast (e.g. by API)
CAP_FAST_SIZE: Final = "fast-size"

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AbstractQuotaModel(metaclass=ABCMeta):
    @abstractmethod
    def mangle_qspath(self, ref: VFolderID | QuotaScopeID | str | None) -> Path:
        raise NotImplementedError

    @abstractmethod
    async def create_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        options: Optional[QuotaConfig] = None,
        extra_args: Optional[dict[str, Any]] = None,
    ) -> None:
        """
        Creates a new quota scope.

        Raises `AlreadyExists` error if there is the quota scope with the same name.
        """
        raise NotImplementedError

    async def get_extra_quota_info(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> dict[str, Any] | None:
        """
        Get the information about the given volume.
        Returns None if target volume does not exist.
        """
        return None

    @abstractmethod
    async def describe_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> Optional[QuotaUsage]:
        """
        Get the information about the given quota scope.
        Returns None if target quota scope does not exist.
        """
        raise NotImplementedError

    @abstractmethod
    async def update_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        config: QuotaConfig,
    ) -> None:
        """
        Update the quota option of the given quota scope.
        """
        raise NotImplementedError

    @abstractmethod
    async def unset_quota(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> None:
        """
        Lifts off quota set on given quota scope.
        """
        raise NotImplementedError

    @abstractmethod
    async def delete_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> None:
        """
        Deletes the given quota scope.

        Raises `NotEmpty` error if there are one or more vfolders inside the quota scope.
        """
        raise NotImplementedError


class AbstractFSOpModel(metaclass=ABCMeta):
    @abstractmethod
    async def copy_tree(
        self,
        src_path: Path,
        dst_path: Path,
    ) -> None:
        """
        The actual backend-specific implementation of copying
        files from a directory to another in an efficient way.
        The source and destination are in the same filesystem namespace
        but they may be on different physical media.
        """
        raise NotImplementedError

    @abstractmethod
    async def move_tree(
        self,
        src_path: Path,
        dst_path: Path,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    async def delete_tree(
        self,
        path: Path,
    ) -> None:
        """
        Deletes all files and directories inside the given path.
        """
        raise NotImplementedError

    @abstractmethod
    def scan_tree(
        self,
        path: Path,
        *,
        recursive: bool = True,
    ) -> AsyncIterator[DirEntry]:
        """
        Iterates over all files within the given path recursively.
        """
        raise NotImplementedError

    @abstractmethod
    async def scan_tree_usage(
        self,
        path: Path,
    ) -> TreeUsage:
        """
        Retrieves the number of bytes and the number of files and directories inside
        the given path, recursively.
        """
        raise NotImplementedError

    @abstractmethod
    async def scan_tree_size(
        self,
        path: Path,
    ) -> BinarySize:
        """
        Retrieves the approximate number of bytes used by a directory,
        including all subdirectories and files recursively.

        This method can be implemented using :meth:`scan_tree_usage()`, but in many cases we can
        often implement this using a faster, dedicated command like ``du``.
        """
        raise NotImplementedError


class AbstractVolume(metaclass=ABCMeta):
    quota_model: AbstractQuotaModel
    fsop_model: AbstractFSOpModel
    name: ClassVar[str] = "undefined"

    def __init__(
        self,
        local_config: Mapping[str, Any],
        mount_path: Path,
        *,
        etcd: AsyncEtcd,
        event_dispatcher: EventDispatcher,
        event_producer: EventProducer,
        watcher: Optional[WatcherClient] = None,
        options: Optional[Mapping[str, Any]] = None,
    ) -> None:
        self.local_config = local_config
        self.mount_path = mount_path
        self.config = options or {}
        self.etcd = etcd
        self.event_dispatcher = event_dispatcher
        self.event_producer = event_producer
        self.watcher = watcher

    async def init(self) -> None:
        self.fsop_model = await self.create_fsop_model()
        self.quota_model = await self.create_quota_model()

    async def shutdown(self) -> None:
        pass

    @abstractmethod
    def info(self) -> VolumeInfo:
        raise NotImplementedError

    @abstractmethod
    async def create_quota_model(self) -> AbstractQuotaModel:
        raise NotImplementedError

    @abstractmethod
    async def create_fsop_model(self) -> AbstractFSOpModel:
        raise NotImplementedError

    @final
    def mangle_vfpath(self, vfid: VFolderID) -> Path:
        folder_id_hex = vfid.folder_id.hex
        prefix1 = folder_id_hex[0:2]
        prefix2 = folder_id_hex[2:4]
        rest = folder_id_hex[4:]
        return self.quota_model.mangle_qspath(vfid.quota_scope_id) / prefix1 / prefix2 / rest

    @final
    def sanitize_vfpath(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath = PurePosixPath("."),
    ) -> Path:
        vfpath = self.mangle_vfpath(vfid).resolve()
        if not (vfpath.exists() and vfpath.is_dir()):
            raise VFolderNotFoundError(f"VFolder not found or not a directory: {vfid}")
        target_path = (vfpath / relpath).resolve()
        if not target_path.is_relative_to(vfpath):
            raise InvalidSubpathError(vfid, relpath)
        return target_path

    @final
    def strip_vfpath(self, vfid: VFolderID, target_path: Path) -> PurePosixPath:
        vfpath = self.mangle_vfpath(vfid).resolve()
        return PurePosixPath(target_path.relative_to(vfpath))

    # ------ volume operations -------

    @abstractmethod
    async def get_capabilities(self) -> FrozenSet[str]:
        raise NotImplementedError

    @abstractmethod
    async def get_hwinfo(self) -> HardwareMetadata:
        raise NotImplementedError

    @abstractmethod
    async def create_vfolder(
        self,
        vfid: VFolderID,
        exist_ok: bool = False,
        mode: int = DEFAULT_VFOLDER_PERMISSION_MODE,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    async def delete_vfolder(self, vfid: VFolderID) -> None:
        raise NotImplementedError

    @abstractmethod
    async def clone_vfolder(
        self,
        src_vfid: VFolderID,
        dst_vfid: VFolderID,
    ) -> None:
        """
        Create a new vfolder on the same volume and copy all contents of the source
        vfolder into it, preserving file permissions and timestamps.
        """
        raise NotImplementedError

    @abstractmethod
    async def get_vfolder_mount(self, vfid: VFolderID, subpath: str) -> Path:
        raise NotImplementedError

    @abstractmethod
    async def put_metadata(self, vfid: VFolderID, payload: bytes) -> None:
        pass

    @abstractmethod
    async def get_metadata(self, vfid: VFolderID) -> bytes:
        pass

    @abstractmethod
    async def get_performance_metric(self) -> FSPerfMetric:
        pass

    @abstractmethod
    async def get_fs_usage(self) -> CapacityUsage:
        pass

    @abstractmethod
    async def get_usage(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath = PurePosixPath("."),
    ) -> TreeUsage:
        pass

    @abstractmethod
    async def get_used_bytes(self, vfid: VFolderID) -> BinarySize:
        pass

    # ------ vfolder operations -------

    @abstractmethod
    def scandir(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        *,
        recursive: bool = True,
    ) -> AsyncIterator[DirEntry]:
        pass

    @abstractmethod
    async def mkdir(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        *,
        parents: bool = False,
        exist_ok: bool = False,
    ) -> None:
        pass

    @abstractmethod
    async def rmdir(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        *,
        recursive: bool = False,
    ) -> None:
        pass

    @abstractmethod
    async def move_file(
        self,
        vfid: VFolderID,
        src: PurePosixPath,
        dst: PurePosixPath,
    ) -> None:
        pass

    @abstractmethod
    async def move_tree(
        self,
        vfid: VFolderID,
        src: PurePosixPath,
        dst: PurePosixPath,
    ) -> None:
        pass

    @abstractmethod
    async def copy_file(
        self,
        vfid: VFolderID,
        src: PurePosixPath,
        dst: PurePosixPath,
    ) -> None:
        pass

    @abstractmethod
    async def prepare_upload(self, vfid: VFolderID) -> str:
        """
        Prepare an upload session by creating a dedicated temporary directory.
        Returns a unique session identifier.
        """
        pass

    @abstractmethod
    async def add_file(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        payload: AsyncIterator[bytes],
    ) -> None:
        pass

    @abstractmethod
    def read_file(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        *,
        chunk_size: int = 0,
    ) -> AsyncIterator[bytes]:
        pass

    @abstractmethod
    async def delete_files(
        self,
        vfid: VFolderID,
        relpaths: Sequence[PurePosixPath],
        *,
        recursive: bool = False,
    ) -> None:
        pass



from __future__ import annotations

import asyncio
import logging
import shutil
import tarfile
import tempfile
from collections.abc import AsyncIterator
from pathlib import Path
from typing import Optional, override

import aiofiles
import aiofiles.os

from ai.backend.common.artifact_storage import AbstractStorage
from ai.backend.common.dto.storage.response import (
    PresignedDownloadObjectResponse,
    PresignedUploadObjectResponse,
    VFSFileInfo,
    VFSFileMetaResponse,
    VFSListFilesResponse,
)
from ai.backend.common.types import StreamReader
from ai.backend.logging import BraceStyleAdapter
from ai.backend.storage.config.unified import VFSStorageConfig
from ai.backend.storage.exception import (
    FileStreamDownloadError,
    FileStreamUploadError,
    NotImplementedAPI,
    ObjectInfoFetchError,
    StorageBucketFileNotFoundError,
)
from ai.backend.storage.utils import normalize_filepath

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore


class VFSFileDownloadServerStreamReader(StreamReader):
    """Stream reader that creates a tar archive of a directory on-the-fly."""

    _file_path: Path
    _chunk_size: int

    def __init__(self, file_path: Path, chunk_size: int) -> None:
        self._file_path = file_path
        self._chunk_size = chunk_size

    @override
    async def read(self) -> AsyncIterator[bytes]:
        """Create a tar archive of the directory and stream it."""
        async with aiofiles.open(self._file_path, "rb") as f:
            bytes_streamed = 0
            while True:
                chunk = await f.read(self._chunk_size)
                if not chunk:
                    break
                bytes_streamed += len(chunk)
                yield chunk

    @override
    def content_type(self) -> Optional[str]:
        return "application/octet-stream"


class VFSDirectoryDownloadServerStreamReader(StreamReader):
    """Stream reader that creates a tar archive of a directory on-the-fly."""

    _directory_path: Path
    _chunk_size: int
    _temp_file: Optional[Path]

    def __init__(self, directory_path: Path, chunk_size: int) -> None:
        self._directory_path = directory_path
        self._chunk_size = chunk_size
        self._temp_file = None

    @override
    async def read(self) -> AsyncIterator[bytes]:
        """Create a tar archive of the directory and stream it."""
        loop = asyncio.get_running_loop()

        # Create temporary file for the tar archive
        with tempfile.NamedTemporaryFile(delete=False, suffix=".tar") as temp_file:
            self._temp_file = Path(temp_file.name)

            # Create tar archive in executor to avoid blocking
            await loop.run_in_executor(
                None, self._create_tar_archive, str(self._directory_path), str(self._temp_file)
            )

        try:
            # Stream the tar file
            async with aiofiles.open(self._temp_file, "rb") as f:
                bytes_streamed = 0
                while True:
                    chunk = await f.read(self._chunk_size)
                    if not chunk:
                        break
                    bytes_streamed += len(chunk)
                    yield chunk
        finally:
            # Clean up temp file
            if self._temp_file and self._temp_file.exists():
                await aiofiles.os.remove(self._temp_file)
                log.debug(f"Cleaned up temp file: {self._temp_file}")

    @override
    def content_type(self) -> Optional[str]:
        return "application/x-tar"

    def _create_tar_archive(self, source_dir: str, tar_path: str) -> None:
        """Create tar archive of directory contents."""
        try:
            log.debug(f"Creating tar archive: {source_dir} -> {tar_path}")
            with tarfile.open(tar_path, "w") as tar:
                tar.add(source_dir, arcname=".", recursive=True)
            log.debug(f"Tar archive created successfully: {tar_path}")
        except Exception as e:
            log.error(f"Failed to create tar archive: {e}")
            raise


class VFSStorage(AbstractStorage):
    """
    VFS (Virtual File System) storage backend that uses host filesystem path
    as storage backend, similar to object storage but for local files.
    """

    _name: str
    _base_path: Path
    _upload_chunk_size: int
    _download_chunk_size: int
    _max_file_size: Optional[int]

    def __init__(self, name: str, cfg: VFSStorageConfig) -> None:
        self._name = name
        base_path = cfg.base_path.resolve()
        if cfg.subpath:
            base_path = base_path / cfg.subpath
        self._base_path = base_path
        # Ensure base path exists
        self._base_path.mkdir(parents=True, exist_ok=True)
        self._upload_chunk_size = cfg.upload_chunk_size
        self._download_chunk_size = cfg.download_chunk_size
        self._max_file_size = cfg.max_file_size

    @property
    def name(self) -> str:
        return self._name

    @property
    def base_path(self) -> Path:
        return self._base_path

    def resolve_path(self, filepath: str) -> Path:
        """
        Resolve relative filepath to absolute path within base_path.
        Prevents path traversal attacks.
        """
        # Normalize the filepath (remove .., ., etc.)
        normalized_path = normalize_filepath(filepath)

        # Resolve to absolute path within base_path
        full_path = (self._base_path / normalized_path).resolve()

        # Ensure the resolved path is within base_path
        try:
            full_path.relative_to(self._base_path)
        except ValueError:
            # TODO: Make exception class
            raise ValueError(f"Path traversal not allowed: {filepath}")

        return full_path

    def _validate_upload_constraints(self, filepath: str, file_size: Optional[int] = None) -> None:
        """Validate upload constraints (file extension, size, read-only mode)."""
        # Check file size
        if self._max_file_size and file_size and file_size > self._max_file_size:
            raise FileStreamUploadError(
                f"File size {file_size} exceeds maximum {self._max_file_size}"
            )

    @override
    async def stream_upload(
        self,
        filepath: str,
        data_stream: StreamReader,
    ) -> None:
        """
        Upload a file to VFS using streaming.

        Args:
            filepath: Path to the file to upload (relative to base_path)
            data_stream: Async iterator of file data chunks
        """
        try:
            # Validate constraints first
            self._validate_upload_constraints(filepath)

            # Resolve target path
            target_path = self.resolve_path(filepath)

            # Create parent directories if needed
            target_path.parent.mkdir(parents=True, exist_ok=True)

            # Stream write to file
            total_size = 0
            async with aiofiles.open(target_path, "wb") as f:
                async for chunk in data_stream.read():
                    if chunk:  # Skip empty chunks
                        await f.write(chunk)
                        total_size += len(chunk)

                        # Check size limit during upload
                        if self._max_file_size and total_size > self._max_file_size:
                            # Clean up partial file
                            await f.close()
                            await aiofiles.os.remove(target_path)
                            raise FileStreamUploadError(
                                f"File size exceeds maximum {self._max_file_size}"
                            )

        except Exception as e:
            raise FileStreamUploadError(f"Upload failed: {str(e)}") from e

    @override
    async def stream_download(self, filepath: str) -> StreamReader:
        """
        Download a file or directory from VFS using streaming.

        Args:
            filepath: Path to the file or directory to download (relative to base_path)

        Returns:
            StreamReader: Stream for reading file data or tar archive for directories
        """
        try:
            target_path = self.resolve_path(filepath)

            if not target_path.exists():
                raise FileStreamDownloadError(f"Path not found: {filepath}")

            if target_path.is_dir():
                # Handle directory download as tar archive
                return VFSDirectoryDownloadServerStreamReader(
                    target_path, self._download_chunk_size
                )
            else:
                return VFSFileDownloadServerStreamReader(target_path, self._download_chunk_size)

        except Exception as e:
            raise FileStreamDownloadError(f"Download failed: {str(e)}") from e

    @override
    async def get_file_info(self, filepath: str) -> VFSFileMetaResponse:
        """
        Get file information.

        Args:
            filepath: Path to the file (relative to base_path)

        Returns:
            VFSFileMetaResponse with file metadata
        """
        try:
            target_path = self.resolve_path(filepath)

            if not target_path.exists():
                raise StorageBucketFileNotFoundError(f"File not found: {filepath}")

            stat_result = await aiofiles.os.stat(target_path)

            # Determine content type based on file extension
            content_type = "application/octet-stream"  # Default
            if target_path.suffix:
                import mimetypes

                guessed_type, _ = mimetypes.guess_type(str(target_path))
                if guessed_type:
                    content_type = guessed_type

            return VFSFileMetaResponse(
                filepath=filepath,
                content_length=stat_result.st_size,
                content_type=content_type,
                last_modified=stat_result.st_mtime,
                created=stat_result.st_ctime,
                is_directory=target_path.is_dir(),
                metadata={
                    "mode": oct(stat_result.st_mode),
                    "uid": str(stat_result.st_uid),
                    "gid": str(stat_result.st_gid),
                },
            )

        except Exception as e:
            raise ObjectInfoFetchError(f"Get file info failed: {str(e)}") from e

    @override
    async def delete_file(self, filepath: str) -> None:
        """
        Delete a file or directory.

        Args:
            filepath: Path to the file/directory to delete (relative to base_path)
        """

        try:
            target_path = self.resolve_path(filepath)

            if not target_path.exists():
                # Silently ignore non-existent files (similar to S3 behavior)
                return

            if target_path.is_file():
                await aiofiles.os.remove(target_path)
            elif target_path.is_dir():
                # Remove directory recursively
                loop = asyncio.get_running_loop()
                await loop.run_in_executor(None, shutil.rmtree, target_path)
            else:
                raise FileStreamUploadError(f"Cannot delete: {filepath}")

        except Exception as e:
            raise FileStreamUploadError(f"Delete failed: {str(e)}") from e

    async def list_directory(self, directory: str) -> list[VFSFileInfo]:
        """
        List files and directories in a directory.

        Args:
            directory: Directory path to list (relative to base_path)

        Returns:
            List of file/directory information
        """
        try:
            target_path = self.resolve_path(directory)

            if not target_path.exists():
                raise StorageBucketFileNotFoundError(f"Directory not found: {directory}")

            if not target_path.is_dir():
                raise FileStreamDownloadError(f"Path is not a directory: {directory}")

            entries = []
            entries_iter = await aiofiles.os.scandir(target_path)
            for entry in entries_iter:
                stat_result = await aiofiles.os.stat(entry.path)
                entries.append(
                    VFSFileInfo(
                        name=entry.name,
                        type="directory" if entry.is_dir() else "file",
                        size=stat_result.st_size if entry.is_file() else None,
                        modified=stat_result.st_mtime,
                        created=stat_result.st_ctime,
                        path=str(Path(directory) / entry.name),
                    )
                )

            return entries

        except Exception as e:
            raise FileStreamDownloadError(f"List directory failed: {str(e)}") from e

    async def list_files_recursive(self, directory: str) -> VFSListFilesResponse:
        """
        Recursively list all files in a directory and its subdirectories.

        Args:
            directory: Directory path to start listing from (relative to base_path)

        Returns:
            VFSListFilesResponse containing all files found recursively
        """
        try:
            target_path = self.resolve_path(directory)

            if not target_path.exists():
                raise StorageBucketFileNotFoundError(f"Directory not found: {directory}")

            if not target_path.is_dir():
                raise FileStreamDownloadError(f"Path is not a directory: {directory}")

            files = await self._collect_files_recursive(target_path, directory)

            return VFSListFilesResponse(files=files)

        except Exception as e:
            raise FileStreamDownloadError(f"List files recursively failed: {str(e)}") from e

    async def _collect_files_recursive(
        self, current_path: Path, relative_base: str
    ) -> list[VFSFileInfo]:
        """
        Helper method to recursively collect files.

        Args:
            current_path: Current absolute path being processed
            relative_base: Relative path from the original base directory

        Returns:
            List of found files
        """
        files: list[VFSFileInfo] = []
        entries_iter = await aiofiles.os.scandir(current_path)
        for entry in entries_iter:
            stat_result = await aiofiles.os.stat(entry.path)

            # Calculate relative path
            if relative_base:
                relative_path = str(Path(relative_base) / entry.name)
            else:
                relative_path = entry.name

            file_info = VFSFileInfo(
                name=entry.name,
                type="directory" if entry.is_dir() else "file",
                size=stat_result.st_size if entry.is_file() else None,
                modified=stat_result.st_mtime,
                created=stat_result.st_ctime,
                path=relative_path,
            )
            files.append(file_info)

            # Recursively process subdirectories
            if entry.is_dir():
                subdirectory_files = await self._collect_files_recursive(
                    Path(entry.path), relative_path
                )
                files.extend(subdirectory_files)

        return files

    async def create_directory(self, directory: str) -> None:
        """
        Create a directory.

        Args:
            directory: Directory path to create (relative to base_path)
        """
        try:
            target_path = self.resolve_path(directory)
            target_path.mkdir(parents=True, exist_ok=True)

        except Exception as e:
            raise FileStreamUploadError(f"Create directory failed: {str(e)}") from e

    async def get_disk_usage(self) -> dict:
        """
        Get disk usage information for the base path.

        Returns:
            Dictionary with capacity and used bytes information
        """
        try:
            stat_result = await aiofiles.os.statvfs(self._base_path)

            total_bytes = stat_result.f_frsize * stat_result.f_blocks
            available_bytes = stat_result.f_frsize * stat_result.f_bavail
            used_bytes = total_bytes - available_bytes

            return {
                "capacity_bytes": total_bytes,
                "used_bytes": used_bytes,
                "available_bytes": available_bytes,
            }

        except Exception as e:
            raise FileStreamDownloadError(f"Get disk usage failed: {str(e)}") from e

    async def generate_presigned_upload_url(self, key: str) -> PresignedUploadObjectResponse:
        raise NotImplementedAPI("VFS storage doesn't support presigned upload URLs")

    async def generate_presigned_download_url(
        self, filepath: str
    ) -> PresignedDownloadObjectResponse:
        raise NotImplementedAPI("VFS storage doesn't support presigned download URLs")



from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.common.bgtask.task.registry import BackgroundTaskHandlerRegistry

from .tasks.clone import VFolderCloneTaskHandler
from .tasks.delete import VFolderDeleteTaskHandler

if TYPE_CHECKING:
    from ai.backend.common.events.dispatcher import EventProducer

    from ..volumes.pool import VolumePool


class BgtaskHandlerRegistryCreator:
    def __init__(self, volume_pool: VolumePool, event_producer: EventProducer) -> None:
        self._volume_pool = volume_pool
        self._event_producer = event_producer

    def create(self) -> BackgroundTaskHandlerRegistry:
        registry = BackgroundTaskHandlerRegistry()
        registry.register(VFolderCloneTaskHandler(self._volume_pool, self._event_producer))
        registry.register(VFolderDeleteTaskHandler(self._volume_pool, self._event_producer))

        return registry



from enum import StrEnum
from typing import Final

import trafaret as t

from ai.backend.common import validators as tx

DEFAULT_CLUSTER_ID: Final = 1
DEFAULT_CLUSTER_INFO_CACHE_TTL: Final = 120.0  # seconds


class APIVersion(StrEnum):
    V1 = "v1"
    V2 = "v2"
    LATEST = "latest"


config_iv = t.Dict({
    t.Key("vast_endpoint"): t.String(),
    t.Key("vast_username"): t.String(),
    t.Key("vast_password"): t.String(),
    t.Key("vast_verify_ssl", default=False): t.ToBool(),
    t.Key("vast_force_login", default=True): t.ToBool(),
    t.Key("vast_api_version", default=APIVersion.V2): tx.Enum(APIVersion),
    t.Key("vast_cluster_id", default=DEFAULT_CLUSTER_ID): t.Int,
    t.Key("vast_storage_base_dir", default="/"): t.String(),
    t.Key("vast_cluster_info_cache_ttl", default=DEFAULT_CLUSTER_INFO_CACHE_TTL): t.Float,
}).allow_extra("*")



from pathlib import PurePosixPath

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.common.types import VFolderID


class StorageProxyError(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/storage/generic"
    error_title = "Storage Proxy Error"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class ProcessExecutionError(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/storage/execution/failed"
    error_title = "Storage Operation Execution Failed"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class ExternalStorageServiceError(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/storage/external/failed"
    error_title = "External Operation Failed"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class NetAppClientError(ExternalStorageServiceError, web.HTTPServiceUnavailable):
    error_type = "https://api.backend.ai/probs/storage/netapp/api-error"
    error_title = "NetApp API Error"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class VFolderNotFoundError(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/storage/vfolder/not-found"
    error_title = "VFolder Not Found"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.VFOLDER,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class QuotaDirectoryNotEmptyError(BackendAIError, web.HTTPConflict):
    error_type = "https://api.backend.ai/probs/storage/quota-directory-not-empty"
    error_title = "Quota Directory Not Empty"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.SOFT_DELETE,
            error_detail=ErrorDetail.CONFLICT,
        )


class QuotaScopeNotFoundError(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/storage/quota/scope/not-found"
    error_title = "Quota Scope Not Found"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.QUOTA_SCOPE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class QuotaScopeAlreadyExists(BackendAIError, web.HTTPConflict):
    error_type = "https://api.backend.ai/probs/storage/quota/scope/already-exists"
    error_title = "Quota Scope Already Exists"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.ALREADY_EXISTS,
        )


class InvalidQuotaConfig(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/storage/quota/config/invalid"
    error_title = "Invalid Quota Config"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.UPDATE,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class InvalidSubpathError(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/storage/subpath/invalid"
    error_title = "Invalid Subpath"

    def __init__(self, vfid: VFolderID, relpath: PurePosixPath) -> None:
        msg_str = f"Invalid Subpath (vfid={vfid}, relpath={relpath})"
        super().__init__(extra_msg=msg_str, extra_data=msg_str)

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class InvalidQuotaScopeError(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/storage/quota/scope/invalid"
    error_title = "Invalid Quota Scope"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class InvalidVolumeError(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/storage/volume/invalid"
    error_title = "Invalid Volume"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class InvalidAPIParameters(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/storage/invalid-api-params"
    error_title = "Invalid API parameters"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class FileStreamUploadError(ProcessExecutionError):
    error_type = "https://api.backend.ai/probs/storage/file-stream-upload-failed"
    error_title = "Failed to upload file stream"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class FileStreamDownloadError(ProcessExecutionError):
    error_type = "https://api.backend.ai/probs/storage/file-stream-download-failed"
    error_title = "Failed to download file stream"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class PresignedUploadURLGenerationError(ProcessExecutionError):
    error_type = "https://api.backend.ai/probs/storage/presigned-upload-url-generation-failed"
    error_title = "Failed to generate presigned upload URL"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class PresignedDownloadURLGenerationError(ProcessExecutionError):
    error_type = "https://api.backend.ai/probs/storage/presigned-download-url-generation-failed"
    error_title = "Failed to generate presigned download URL"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class ObjectInfoFetchError(ProcessExecutionError):
    error_type = "https://api.backend.ai/probs/storage/object-info-fetch-failed"
    error_title = "Failed to fetch object info"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class StorageNotFoundError(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/storage/object-not-found"
    error_title = "Storage Config Not Found"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class StorageTypeInvalidError(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/storage/object-type-invalid"
    error_title = "Storage Config Invalid Type"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class ObjectStorageBucketNotFoundError(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/storage/bucket/object-not-found"
    error_title = "Storage Bucket Not Found"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class StorageBucketFileNotFoundError(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/storage/bucket/file/object-not-found"
    error_title = "Storage Bucket File Not Found"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class RegistryNotFoundError(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/registries/registry-not-found"
    error_title = "Registry Not Found"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ARTIFACT_REGISTRY,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class HuggingFaceAPIError(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/registries/huggingface/api-error"
    error_title = "HuggingFace API Error"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ARTIFACT_REGISTRY,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class HuggingFaceModelNotFoundError(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/registries/huggingface/model-not-found"
    error_title = "HuggingFace Model Not Found"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ARTIFACT,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class ObjectStorageConfigInvalidError(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/storage/object/config/invalid"
    error_title = "Object Storage Config Invalid"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.UPDATE,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class ReservoirStorageConfigInvalidError(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/storage/reservoir/config/invalid"
    error_title = "Reservoir Storage Config Invalid"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.UPDATE,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class ArtifactStorageEmptyError(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/storage/artifact/config/invalid"
    error_title = "Artifact Storage Empty"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ARTIFACT,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class ArtifactRevisionEmptyError(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/storage/artifact/revision/empty"
    error_title = "Artifact Revision Empty"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ARTIFACT,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class ArtifactImportError(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/storage/artifact/import/failed"
    error_title = "Artifact Import Failed"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.ARTIFACT,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class NotImplementedAPI(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/storage/api/not-implemented"
    error_title = "API Not Implemented"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.NOT_IMPLEMENTED,
        )


class ObjectStorageObjectDeletionError(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/storage/object/deletion/failed"
    error_title = "Object Deletion Failed"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.OBJECT_STORAGE,
            operation=ErrorOperation.HARD_DELETE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class StorageTransferError(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/storage/transfer/failed"
    error_title = "Storage Transfer Failed"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class StorageStepRequiredStepNotProvided(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/storage/step-mapping-not-provided"
    error_title = "Storage Step Mapping Not Provided"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class ArtifactVerifyStorageTypeInvalid(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/storage/artifact/verify/storage-type/invalid"
    error_title = "Artifact Verify Storage Type Invalid"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class ArtifactVerificationFailedError(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/storage/artifact/verification/failed"
    error_title = "Artifact Verification Failed"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.STORAGE_PROXY,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.BAD_REQUEST,
        )



from __future__ import annotations

import pathlib
from typing import Optional

import click
from setproctitle import setproctitle

from ai.backend.common.cli import LazyGroup
from ai.backend.logging import LogLevel

from .context import CLIContext


@click.group(invoke_without_command=False, context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        exists=True,
        path_type=pathlib.Path,
    ),
    default=None,
    help="The config file path. (default: ./manager.conf and /etc/backend.ai/manager.conf)",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Set the logging level to DEBUG",
)
@click.option(
    "--log-level",
    type=click.Choice([*LogLevel], case_sensitive=False),
    default=LogLevel.NOTSET,
    help="Set the logging verbosity level",
)
@click.pass_context
def main(
    ctx: click.Context,
    log_level: LogLevel,
    debug: bool,
    config_path: Optional[pathlib.Path] = None,
) -> None:
    """The root entrypoint for unified CLI of storage-proxy"""
    setproctitle("backend.ai: storage.cli")
    if debug:
        log_level = LogLevel.DEBUG

    ctx.obj = ctx.with_resource(CLIContext(config_path=config_path, log_level=log_level))


@main.group(cls=LazyGroup, import_name="ai.backend.storage.cli.config:cli")
def config() -> None:
    """Command set for configuration management."""



from __future__ import annotations

import logging
from http import HTTPStatus
from typing import TYPE_CHECKING
from urllib.parse import unquote

from aiohttp import web

from ai.backend.common.api_handlers import (
    APIResponse,
    BodyParam,
    PathParam,
    QueryParam,
    api_handler,
)
from ai.backend.common.data.artifact.types import ArtifactRegistryType
from ai.backend.common.data.storage.registries.types import ModelTarget
from ai.backend.common.dto.storage.request import (
    HuggingFaceGetCommitHashReqPathParam,
    HuggingFaceGetCommitHashReqQueryParam,
    HuggingFaceImportModelsReq,
    HuggingFaceRetrieveModelReqPathParam,
    HuggingFaceRetrieveModelReqQueryParam,
    HuggingFaceRetrieveModelsReq,
    HuggingFaceScanModelsReq,
)
from ai.backend.common.dto.storage.response import (
    HuggingFaceGetCommitHashResponse,
    HuggingFaceImportModelsResponse,
    HuggingFaceRetrieveModelResponse,
    HuggingFaceRetrieveModelsResponse,
    HuggingFaceScanModelsResponse,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.storage.config.unified import (
    HuggingfaceConfig,
    LegacyHuggingfaceConfig,
)
from ai.backend.storage.services.artifacts.huggingface import (
    HuggingFaceService,
    HuggingFaceServiceArgs,
    create_huggingface_import_pipeline,
)

from ....utils import log_client_api_entry

if TYPE_CHECKING:
    from ....context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class HuggingFaceRegistryAPIHandler:
    _huggingface_service: HuggingFaceService

    def __init__(self, huggingface_service: HuggingFaceService) -> None:
        self._huggingface_service = huggingface_service

    @api_handler
    async def scan_models(
        self,
        body: BodyParam[HuggingFaceScanModelsReq],
    ) -> APIResponse:
        """
        Scan HuggingFace registry and return metadata.
        """
        await log_client_api_entry(log, "scan_models", body.parsed)

        models = await self._huggingface_service.scan_models(
            registry_name=body.parsed.registry_name,
            limit=body.parsed.limit,
            search=body.parsed.search,
            sort=body.parsed.order,
        )

        response = HuggingFaceScanModelsResponse(
            models=models,
        )

        return APIResponse.build(
            status_code=HTTPStatus.OK,
            response_model=response,
        )

    @api_handler
    async def retrieve_model(
        self,
        path: PathParam[HuggingFaceRetrieveModelReqPathParam],
        query: QueryParam[HuggingFaceRetrieveModelReqQueryParam],
    ) -> APIResponse:
        """
        Retrieve HuggingFace registry model data.
        """
        await log_client_api_entry(log, "retrieve_model", path.parsed.model_id)

        model_id = unquote(path.parsed.model_id)
        model_data = await self._huggingface_service.retrieve_model(
            registry_name=query.parsed.registry_name,
            model=ModelTarget(
                model_id=model_id,
                revision=query.parsed.revision,
            ),
        )

        response = HuggingFaceRetrieveModelResponse(
            model=model_data,
        )

        return APIResponse.build(
            status_code=HTTPStatus.OK,
            response_model=response,
        )

    @api_handler
    async def retrieve_models(
        self,
        body: BodyParam[HuggingFaceRetrieveModelsReq],
    ) -> APIResponse:
        """
        Retrieve HuggingFace registry model data.
        """
        await log_client_api_entry(log, "retrieve_models", body.parsed)

        model_datas = await self._huggingface_service.retrieve_models(
            registry_name=body.parsed.registry_name,
            models=body.parsed.models,
        )

        response = HuggingFaceRetrieveModelsResponse(
            models=model_datas,
        )

        return APIResponse.build(
            status_code=HTTPStatus.OK,
            response_model=response,
        )

    @api_handler
    async def scan_models_sync(
        self,
        body: BodyParam[HuggingFaceScanModelsReq],
    ) -> APIResponse:
        """
        Scan HuggingFace registry and return metadata including README content synchronously.
        This endpoint waits for all metadata (including README) to be fully downloaded before responding.
        """
        await log_client_api_entry(log, "scan_models_sync", body.parsed)

        models = await self._huggingface_service.scan_models_sync(
            registry_name=body.parsed.registry_name,
            limit=body.parsed.limit,
            search=body.parsed.search,
            sort=body.parsed.order,
        )

        response = HuggingFaceScanModelsResponse(
            models=models,
        )

        return APIResponse.build(
            status_code=HTTPStatus.OK,
            response_model=response,
        )

    @api_handler
    async def get_commit_hash(
        self,
        path: PathParam[HuggingFaceGetCommitHashReqPathParam],
        query: QueryParam[HuggingFaceGetCommitHashReqQueryParam],
    ) -> APIResponse:
        """
        Get the commit hash for a specific HuggingFace model revision.
        """
        await log_client_api_entry(log, "get_commit_hash", path.parsed.model_id)

        model_id = unquote(path.parsed.model_id)
        commit_hash = await self._huggingface_service.get_model_commit_hash(
            registry_name=query.parsed.registry_name,
            model=ModelTarget(
                model_id=model_id,
                revision=query.parsed.revision,
            ),
        )

        response = HuggingFaceGetCommitHashResponse(
            commit_hash=commit_hash,
        )

        return APIResponse.build(
            status_code=HTTPStatus.OK,
            response_model=response,
        )

    @api_handler
    async def import_models(
        self,
        body: BodyParam[HuggingFaceImportModelsReq],
    ) -> APIResponse:
        """
        Import multiple HuggingFace models to storage in batch.
        """
        await log_client_api_entry(log, "import_models", body.parsed)

        # Create import pipeline based on storage step mappings
        pipeline = create_huggingface_import_pipeline(
            registry_configs=self._huggingface_service._registry_configs,
            transfer_manager=self._huggingface_service._transfer_manager,
            storage_step_mappings=body.parsed.storage_step_mappings,
            artifact_verifier_ctx=self._huggingface_service._artifact_verifier_ctx,
            event_producer=self._huggingface_service._event_producer,
            redis_client=self._huggingface_service._redis_client,
        )

        task_id = await self._huggingface_service.import_models_batch(
            registry_name=body.parsed.registry_name,
            models=body.parsed.models,
            storage_step_mappings=body.parsed.storage_step_mappings,
            pipeline=pipeline,
        )

        response = HuggingFaceImportModelsResponse(
            task_id=task_id,
        )

        return APIResponse.build(
            status_code=HTTPStatus.ACCEPTED,
            response_model=response,
        )


def create_app(ctx: RootContext) -> web.Application:
    app = web.Application()
    app["ctx"] = ctx
    app["prefix"] = "v1/registries/huggingface"

    # Get huggingface configs from new artifact_registries
    huggingface_registry_configs: dict[str, HuggingfaceConfig] = {
        name: r.huggingface
        for name, r in ctx.local_config.artifact_registries.items()
        if r.registry_type == ArtifactRegistryType.HUGGINGFACE and r.huggingface is not None
    }

    # Legacy registry configs support - add from legacy registries for backward compatibility
    for legacy_registry in ctx.local_config.registries:
        if isinstance(legacy_registry.config, LegacyHuggingfaceConfig):
            huggingface_registry_configs[legacy_registry.name] = legacy_registry.config

    huggingface_service = HuggingFaceService(
        HuggingFaceServiceArgs(
            background_task_manager=ctx.background_task_manager,
            storage_pool=ctx.storage_pool,
            registry_configs=huggingface_registry_configs,
            event_producer=ctx.event_producer,
            artifact_verifier_ctx=ctx.artifact_verifier_ctx,
            redis_client=ctx.valkey_artifact_client,
        )
    )
    huggingface_api_handler = HuggingFaceRegistryAPIHandler(huggingface_service=huggingface_service)

    app.router.add_route("POST", "/scan", huggingface_api_handler.scan_models)
    app.router.add_route("POST", "/scan/sync", huggingface_api_handler.scan_models_sync)
    app.router.add_route("POST", "/import", huggingface_api_handler.import_models)

    app.router.add_route("GET", "/model/{model_id}", huggingface_api_handler.retrieve_model)
    app.router.add_route(
        "GET", "/model/{model_id}/commit-hash", huggingface_api_handler.get_commit_hash
    )
    app.router.add_route("POST", "/models/batch", huggingface_api_handler.retrieve_models)
    return app



import asyncio
import logging
import os
from pathlib import Path
from stat import S_IFDIR, S_IFLNK
from subprocess import CalledProcessError
from typing import AsyncIterator

from ai.backend.common.json import load_json
from ai.backend.logging.utils import BraceStyleAdapter

from ...subproc import run
from ...types import DirEntry, DirEntryType, Stat, TreeUsage
from ...utils import fstime2datetime
from .rapidfiles import RapidFileToolsFSOpModel

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class RapidFileToolsv2FSOpModel(RapidFileToolsFSOpModel):
    async def copy_tree(
        self,
        src_path: Path,
        dst_path: Path,
    ) -> None:
        extra_opts: list[bytes] = []
        if src_path.is_dir():
            extra_opts.append(b"-r")
        if dst_path.is_dir():
            extra_opts.append(b"-T")
        try:
            await run([  # noqa: F821
                b"pcopy",
                *extra_opts,
                b"-p",
                # os.fsencode(src_path / "."),  # TODO: check if "/." is necessary?
                os.fsencode(src_path),
                os.fsencode(dst_path),
            ])
        except CalledProcessError as e:
            raise RuntimeError(f'"pcopy" command failed: {e.stderr}')

    def scan_tree(
        self,
        path: Path,
        *,
        recursive: bool = True,
    ) -> AsyncIterator[DirEntry]:
        raw_target_path = os.fsencode(path)

        async def _aiter() -> AsyncIterator[DirEntry]:
            proc = await asyncio.create_subprocess_exec(
                b"pls",
                b"--jsonlines",
                raw_target_path,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.STDOUT,
            )
            assert proc.stdout is not None
            try:
                while True:
                    line = await proc.stdout.readline()
                    if not line:
                        break
                    line = line.rstrip(b"\n")
                    item = load_json(line)

                    item_path = Path(item["path"])
                    entry_type = DirEntryType.FILE
                    if item["filetype"] == S_IFDIR:
                        entry_type = DirEntryType.DIRECTORY
                    if item["filetype"] == S_IFLNK:
                        entry_type = DirEntryType.SYMLINK
                    yield DirEntry(
                        name=item_path.name,
                        path=item_path,
                        type=entry_type,
                        stat=Stat(
                            size=item["size"],
                            owner=str(item["uid"]),
                            mode=item["mode"],
                            modified=fstime2datetime(item["mtime"]),
                            created=fstime2datetime(item["ctime"]),
                        ),
                        symlink_target="",  # TODO: should be tested on PureStorage
                    )
            finally:
                await proc.wait()

        return _aiter()

    async def scan_tree_usage(
        self,
        path: Path,
    ) -> TreeUsage:
        total_size = 0
        total_count = 0
        raw_target_path = os.fsencode(path)
        # Measure the exact file sizes and bytes
        proc = await asyncio.create_subprocess_exec(
            b"pdu",
            b"-0",
            b"-b",
            b"-a",
            raw_target_path,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
        )
        assert proc.stdout is not None
        try:
            # TODO: check slowdowns when there are millions of files
            while True:
                try:
                    line = await proc.stdout.readuntil(b"\0")
                    line = line.rstrip(b"\0")
                except asyncio.IncompleteReadError:
                    break
                size, name = line.split(maxsplit=1)
                if len(name) != len(raw_target_path) and name != raw_target_path:
                    total_size += int(size)
                    total_count += 1
        finally:
            await proc.wait()
        return TreeUsage(file_count=total_count, used_bytes=total_size)



from collections.abc import Sequence
from datetime import datetime
from pathlib import Path, PurePosixPath
from typing import Any, AsyncIterator, Optional, override

from ai.backend.common.defs import DEFAULT_VFOLDER_PERMISSION_MODE, NOOP_STORAGE_BACKEND_TYPE
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events.dispatcher import EventDispatcher, EventProducer
from ai.backend.common.types import BinarySize, HardwareMetadata, QuotaScopeID

from ...types import (
    CapacityUsage,
    DirEntry,
    DirEntryType,
    FSPerfMetric,
    QuotaConfig,
    QuotaUsage,
    Stat,
    TreeUsage,
    VFolderID,
    VolumeInfo,
)
from ..abc import AbstractFSOpModel, AbstractQuotaModel, AbstractVolume


async def _return_empty_dir_entry() -> AsyncIterator[DirEntry]:
    yield DirEntry(
        "", Path(), DirEntryType.FILE, Stat(0, "", 0, datetime.now(), datetime.now()), ""
    )


class NoopQuotaModel(AbstractQuotaModel):
    def __init__(self) -> None:
        pass

    def mangle_qspath(self, ref: VFolderID | QuotaScopeID | str | None) -> Path:
        return Path()

    async def create_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        options: Optional[QuotaConfig] = None,
        extra_args: Optional[dict[str, Any]] = None,
    ) -> None:
        pass

    async def describe_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> Optional[QuotaUsage]:
        pass

    async def update_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        config: QuotaConfig,
    ) -> None:
        pass

    async def unset_quota(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> None:
        pass

    async def delete_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> None:
        pass


class NoopFSOpModel(AbstractFSOpModel):
    def __init__(self) -> None:
        return

    async def copy_tree(
        self,
        src_path: Path,
        dst_path: Path,
    ) -> None:
        pass

    async def move_tree(
        self,
        src_path: Path,
        dst_path: Path,
    ) -> None:
        pass

    async def delete_tree(
        self,
        path: Path,
    ) -> None:
        pass

    def scan_tree(
        self,
        path: Path,
        *,
        recursive: bool = True,
    ) -> AsyncIterator[DirEntry]:
        return _return_empty_dir_entry()

    async def scan_tree_usage(
        self,
        path: Path,
    ) -> TreeUsage:
        return TreeUsage(0, 0)

    async def scan_tree_size(
        self,
        path: Path,
    ) -> BinarySize:
        return BinarySize(0)


class NoopVolume(AbstractVolume):
    name = NOOP_STORAGE_BACKEND_TYPE

    @override
    def info(self) -> VolumeInfo:
        return VolumeInfo(
            backend=NOOP_STORAGE_BACKEND_TYPE,
            path=self.mount_path,
            fsprefix=None,
            options=None,
        )

    async def create_quota_model(self) -> AbstractQuotaModel:
        return NoopQuotaModel()

    async def create_fsop_model(self) -> AbstractFSOpModel:
        return NoopFSOpModel()

    # ------ volume operations -------

    async def get_capabilities(self) -> frozenset[str]:
        return frozenset()

    async def get_hwinfo(self) -> HardwareMetadata:
        return {
            "status": "healthy",
            "status_info": None,
            "metadata": {},
        }

    async def create_vfolder(
        self,
        vfid: VFolderID,
        exist_ok: bool = False,
        mode: int = DEFAULT_VFOLDER_PERMISSION_MODE,
    ) -> None:
        return None

    async def delete_vfolder(self, vfid: VFolderID) -> None:
        return None

    async def clone_vfolder(
        self,
        src_vfid: VFolderID,
        dst_vfid: VFolderID,
    ) -> None:
        return None

    async def get_vfolder_mount(self, vfid: VFolderID, subpath: str) -> Path:
        return Path()

    async def put_metadata(self, vfid: VFolderID, payload: bytes) -> None:
        pass

    async def get_metadata(self, vfid: VFolderID) -> bytes:
        return b""

    async def get_performance_metric(self) -> FSPerfMetric:
        return FSPerfMetric(0, 0, 0, 0, 0.0, 0.0)

    async def get_fs_usage(self) -> CapacityUsage:
        return CapacityUsage(0, 0)

    async def get_usage(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath = PurePosixPath("."),
    ) -> TreeUsage:
        return TreeUsage(0, 0)

    async def get_used_bytes(self, vfid: VFolderID) -> BinarySize:
        return BinarySize(0)

    # ------ vfolder operations -------

    def scandir(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        *,
        recursive: bool = True,
    ) -> AsyncIterator[DirEntry]:
        return _return_empty_dir_entry()

    async def mkdir(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        *,
        parents: bool = False,
        exist_ok: bool = False,
    ) -> None:
        pass

    async def rmdir(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        *,
        recursive: bool = False,
    ) -> None:
        pass

    async def move_file(
        self,
        vfid: VFolderID,
        src: PurePosixPath,
        dst: PurePosixPath,
    ) -> None:
        pass

    async def move_tree(
        self,
        vfid: VFolderID,
        src: PurePosixPath,
        dst: PurePosixPath,
    ) -> None:
        pass

    async def copy_file(
        self,
        vfid: VFolderID,
        src: PurePosixPath,
        dst: PurePosixPath,
    ) -> None:
        pass

    async def prepare_upload(self, vfid: VFolderID) -> str:
        return ""

    async def add_file(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        payload: AsyncIterator[bytes],
    ) -> None:
        pass

    def read_file(
        self,
        vfid: VFolderID,
        relpath: PurePosixPath,
        *,
        chunk_size: int = 0,
    ) -> AsyncIterator[bytes]:
        async def _noop() -> AsyncIterator[bytes]:
            yield b""

        return _noop()

    async def delete_files(
        self,
        vfid: VFolderID,
        relpaths: Sequence[PurePosixPath],
        *,
        recursive: bool = False,
    ) -> None:
        pass


def init_noop_volume(
    etcd: AsyncEtcd,
    event_dispatcher: EventDispatcher,
    event_producer: EventProducer,
) -> NoopVolume:
    return NoopVolume(
        {},
        Path(),
        etcd=etcd,
        event_dispatcher=event_dispatcher,
        event_producer=event_producer,
        watcher=None,
        options=None,
    )
